# import dependencies
import os
import sys
import mutagen
import io
import random
import socket
import re
from functools import partial
from PIL import Image as PImage

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

# import all the other kivy stuff
from kivy.app import App
from kivy.logger import Logger
from kivy.graphics import Rectangle,Color
from kivy.core.image import Image as CoreImage
from kivy.clock import Clock
from kivy.uix.tabbedpanel import TabbedPanel
from kivy.uix.button import Button
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.image import Image
from kivy.uix.popup import Popup
from kivy.factory import Factory
from kivy.support import install_twisted_reactor
from kivy.properties import ObjectProperty

# import our local modules
from kmpc.extra import KmpcHelpers
from kmpc.playlistpanel import PlaylistTabbedPanelItem
from kmpc.mpdfactory import MpdConnection
from kmpc.widgets import InfoLargeLabel,CoverButton,ImageButton,ExtraSlider,ClearButton,OutlineLabel,normalfont,fontawesomefont,backdrop,clearimage,ratingstars

# set the maximum size for cover images, to prevent texture overloading
max_cover_size=1000

Helpers=KmpcHelpers()

class KmpcInterface(TabbedPanel):
    """The main class that ties it all together."""
    mpdconnection=ObjectProperty(None)
    syncmpdconnection=ObjectProperty(None)

    def __init__(self,config,**kwargs):
        """Zero out variables, pull in config file, connect to mpd."""
        super(self.__class__,self).__init__(**kwargs)
        # bind callbacks for tab changes
        self.bind(current_tab=self.main_tab_changed)
        self.mpd_status={'state':'stop','repeat':0,'single':0,'random':0,'consume':0,'curpos':0}
        self.currsong=None
        self.nextsong=None
        self.currfile=None
        self.track_slider_task=None
        self.accessoryPopup=Factory.AccessoryPopup()
        self.tcolor=1
        self.ocolor=0
        self.settingsPopup=Factory.SettingsPopup()
        self.config=config
        self.do_idle_handler=True
        # install twisted reactor to interface with mpd
        install_twisted_reactor()
        # open mpd connection
        self.mpdconnection=MpdConnection(self.config,self.config.get('mpd','mpdhost'),self.config.get('mpd','mpdport'),self.mpd_idle_handler,[self.init_mpd])

    def settings_popup(self):
        self.settingsPopup.open()
        # get the host's IP address and display it
        self.settingsPopup.ids.ip_label.text="IP Address: "+format(self.get_ip())
        self.mpdconnection.protocol.status().addCallback(partial(self.update_mixers,self.settingsPopup)).addErrback(self.handle_mpd_error)
        self.mpdconnection.protocol.replay_gain_status().addCallback(partial(self.update_replaygain,self.settingsPopup)).addErrback(self.handle_mpd_error)

    def update_replaygain(self,p,result):
        rg=str(result)
        p.ids['rg_'+rg].state='down'

    def update_mixers(self,p,result):
        # set up the crossfade slider
        if 'xfade' in result:
            v = int(result['xfade'])
        else:
            v = 0
        p.ids.crossfade_slider.value=v
        # set up the mixrampdb slider
        if 'mixrampdb' in result:
            v = round(float(result['mixrampdb']),6)
        else:
            v = 0.0
        p.ids.mixrampdb_slider.value=float(str(v)[1:])
        # set up the mixrampdelay slider
        if 'mixrampdelay' in result:
            v = round(float(result['mixrampdelay']),6)
        else:
            v = 0.0
        p.ids.mixrampdelay_slider.value=v

    def change_replaygain(self,v):
        Logger.debug("NowPlaying: change_replaygain to "+format(v))
        self.mpdconnection.protocol.replay_gain_mode(str(v)).addErrback(self.handle_mpd_error)

    def change_text_color(self,color):
        Logger.debug("NowPlaying: change_text_color to "+format(color))
        self.tcolor=color
        def _tc(widget):
            for child in widget.children:
                _tc(child)
            if issubclass(widget.__class__,OutlineLabel):
                widget.color=[color,color,color,1]
        for child in self.children:
            _tc(child)

    def change_outline_color(self,color):
        Logger.debug("NowPlaying: change_outline_color to "+format(color))
        self.ocolor=color
        def _tc(widget):
            for child in widget.children:
                _tc(child)
            if issubclass(widget.__class__,OutlineLabel):
                widget.outline_color=[color,color,color,1]
        for child in self.children:
            _tc(child)

    def change_crossfade(self,v):
        """Callback when user changes crossfade slider."""
        Logger.info('Settings: change_crossfade')
        self.mpdconnection.protocol.crossfade(str(v)).addErrback(self.handle_mpd_error)

    def change_mixrampdb(self,v):
        """Callback when user changes mixrampdb slider."""
        Logger.info('Settings: change_mixrampdb')
        self.mpdconnection.protocol.mixrampdb(str(0.0-v)).addErrback(self.handle_mpd_error)

    def change_mixrampdelay(self,v):
        """Callback when user changes mixrampdelay slider."""
        Logger.info('Settings: change_mixrampdelay')
        self.mpdconnection.protocol.mixrampdelay(str(v)).addErrback(self.handle_mpd_error)

    def get_ip(self):
        """Method that tries to get the local IP address, and returns localhost if there isn't one."""
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except:
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP

    def main_tab_changed(self,obj,value):
        """Callback when top tab is changed."""
        self.active_tab = value.text
        Logger.info("Application: Changed active tab to "+self.active_tab)
        if self.active_tab != 'Now Playing':
            # pause the track slider task
            self.track_slider_task.cancel()
        else:
            # update current track status
            self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.mpdconnection.handle_mpd_error)
        if self.active_tab == 'Playlist':
            # switching to the playlist tab repopulates it if it is empty
            if len(self.ids.playlist_tab.rv.data) == 0:
                self.mpdconnection.protocol.playlistinfo().addCallback(self.ids.playlist_tab.populate_playlist).addErrback(self.ids.playlist_tab.handle_mpd_error)

    def handle_mpd_error(self,result):
        self.mpdconnection.handle_mpd_error(result)

    def handle_mpd_message(self,result):
        """Callback for mpd 'kmpc' messages."""
        # result is an iterator although it only generally has one member
        for m in result:
            message = m['message']
            Logger.info("Application: MPD Message: "+message)
            if message == 'AccessoryOff':
                # generated by mausberry circuit, accessory power is off
                self.accessoryPopup.open()
            elif message == 'AccessoryOn':
                # generated by mausberry circuit, accessory power is on
                self.accessoryPopup.dismiss()

    def current_track_slider_release(self):
        """Callback when track slider is released."""
        # get the value of the track slider
        curpos = int(self.ids.current_track_slider.value)
        Logger.debug("Application: touch up on track slider at "+str(curpos))
        # seek to that position in the track
        self.mpdconnection.protocol.seekcur(str(curpos)).addErrback(self.handle_mpd_error)

    def current_track_slider_down(self):
        """Callback when track slider is pressed."""
        Logger.debug("Application: touch down on track slider")
        # cancel the once-per-second update while finger is down so it doesn't move out from under it
        self.track_slider_task.cancel()

    def update_track_slider(self,dt):
        """Increment the slider by 1 second every second."""
        Logger.debug("Application: update_track_slider")
        curpos=int(self.ids.current_track_slider.value)+1
        self.ids.current_track_slider.value=curpos

    def stop_zero_stuff(self):
        """Clear the screen and variables if playback is fully stopped."""
        self.ids.current_track_slider.value=0
        self.ids.current_track_slider.max=0
        self.ids.current_playlist_track_number_label.text=''
        self.ids.next_song_artist_label.text = ''
        self.currfile = None
        self.currsong = None
        self.nextsong = None
        self.ids.song_star_button.text=''
        self.ids.song_star_button.disabled=True
        self.ids.album_cover_layout.clear_widgets()
        self.ids.trackinfo.clear_widgets()
        lbl = OutlineLabel(text="Playback Stopped")
        self.ids.trackinfo.add_widget(lbl)
        self.ids.player.canvas.before.clear()
        self.ids.releasetypelabel.text=''
        self.ids.yearlabel.text=''
        self.ids.remasterlabel.text=''

    def update_mpd_status(self,result):
        """Callback when mpd status changes."""
        Logger.debug('NowPlaying: update_mpd_status()')
        # probably state is the only one necessary, but hey
        self.mpd_status['state']=result['state']
        self.mpd_status['repeat']=result['repeat']
        self.mpd_status['single']=result['single']
        self.mpd_status['random']=result['random']
        self.mpd_status['consume']=result['consume']
        if self.mpd_status['state'] == 'stop':
            # if stopped, there are no current or next songs
            self.currsong=None
            self.nextsong=None
        else:
            # save current song, this is a 0-based index into the playlist
            self.currsong=result['song']
            # ask mpd for current song data
            self.mpdconnection.protocol.currentsong().addCallback(partial(self.update_mpd_currentsong,False)).addErrback(self.handle_mpd_error)
            # save next song, this is a 0-based index into the playlist
            if 'nextsong' in result:
                self.nextsong=result['nextsong']
                # ask mpd for next song data
                self.mpdconnection.protocol.playlistinfo(self.nextsong).addCallback(self.update_mpd_nextsong).addErrback(self.handle_mpd_error)
            else:
                self.nextsong=None
        stflags=['normal','down']
        # check various flag states
        self.ids.repeat_button.state=stflags[int(self.mpd_status['repeat'])]
        self.ids.single_button.state=stflags[int(self.mpd_status['single'])]
        self.ids.random_button.state=stflags[int(self.mpd_status['random'])]
        self.ids.consume_button.state=stflags[int(self.mpd_status['consume'])]
        if self.mpd_status['state']=='pause' or self.mpd_status['state']=='stop':
            # play/pause button should be a play button that is unpressed
            self.ids.play_button.state='normal'
            self.ids.play_button.text=u"\uf04b"
            self.track_slider_task.cancel()
        else:
            # play/pause button should be a pause button that is pressed
            self.ids.play_button.state='down'
            self.ids.play_button.text=u"\uf04c"
        if self.mpd_status['state'] == 'stop':
            # zero everything out if we are stopped
            self.stop_zero_stuff()
        else:
            # mpd returns {elapsed seconds}:{total seconds}, the following splits each to minute:second
            c,t=result['time'].split(":")
            # set the max slider value to the total seconds
            self.ids.current_track_slider.max = int(t)
            # set the current slider value to the current seconds
            self.mpd_status['curpos']=int(c)
            self.ids.current_track_slider.value = int(c)
            if self.mpd_status['state']!='pause':
                self.track_slider_task()
            # throws an exception if i don't do this
            a=int(result['song'])+1
            b=int(result['playlistlength'])
            self.ids.current_playlist_track_number_label.text = "%d of %d" % (a,b)
        # update the playlist tab with status results so that current track will be highlighted
        self.ids.playlist_tab.update_mpd_status(result)

    def change_artist_image(self,img,al_path,instance):
        """Called when you click on an artist logo, changes it to another at random."""
        Logger.debug("change_artist_image: (current path is "+img.source+")")
        img_path=img.source
        if len(os.listdir(al_path))>1:
            while img_path==img.source:
                img_path=os.path.join(al_path,random.choice(os.listdir(al_path)))
            if os.path.isfile(img_path):
                img.source=img_path
            Logger.debug("change_artist_image: (new path is "+img.source+")")
        else:
            Logger.debug("change_artist_image: no other choices for logo")

    def update_mpd_currentsong(self,force,result):
        """Callback for mpd currentsong data."""
        Logger.debug('NowPlaying: update_mpd_currentsong()')
        # this is so expensive screen updates only happen if the song has changed since the last time this callback was called
        songchange=False
        # ti is the track info widget
        ti=self.ids.trackinfo
        # if result is undefined, there's not actually a song playing
        if result:
            # if class's current file doesn't match what mpd returns, the song has changed
            if self.currfile != result['file']:
                songchange=True
            # update class's current file
            self.currfile = result['file']
            if songchange or force:
                # clear the track info widget
                ti.clear_widgets()
                # clear the album cover
                self.ids.album_cover_layout.clear_widgets()
                # get the stored star rating
                self.mpdconnection.protocol.sticker_get('song',self.currfile,'rating').addCallback(self.update_mpd_sticker_rating).addErrback(self.handle_mpd_no_sticker)
                # figure out the full path of the file
                bp=self.config.get('paths','musicpath')
                # p is the absolute path
                p=os.path.join(bp,result['file'])
                haslogo=False
                # set the release year if mpd has it
                year=None
                if 'date' in result:
                    year=result['date'][:4]
                # pull the fanart folder from ini file
                fa_path=self.config.get('paths','fanartpath')
                # try to get the artistid, set it to 0000 if it doesn't exist
                try:
                    aids=str(result['musicbrainz_artistid'])
                except KeyError:
                    aids='0000'
                # if a track has multiple artists, split it up
                artistids=aids.split('/')
                # create a widget to hold the cover art
                cbl = BoxLayout(size_hint=(1,1),orientation='horizontal')
                # loop through artistids
                for mb_aid in artistids:
                    # try to find a logo for this artist in the fanart folder
                    try:
                        # look in the 'logo' subfolder of the artistid path
                        al_path=os.path.join(fa_path,mb_aid,'logo')
                        # pick one at random
                        img_path=os.path.join(al_path,random.choice(os.listdir(al_path)))
                        if os.path.isfile(img_path):
                            # create an image button out of the logo so you can press it
                            img = ImageButton(source=os.path.join(al_path,img_path),allow_stretch=True,color=(1,1,1,0.65))
                            img.bind(on_press=partial(self.change_artist_image,img,al_path))
                            cbl.add_widget(img)
                            haslogo=True
                    except OSError:
                        Logger.debug("update_mpd_currentsong: No logos for artist "+mb_aid)
                    except Exception as e:
                        Logger.exception("update_mpd_currentsong: "+format(e))
                if haslogo:
                    # we found a logo, show it
                    current_artist_label = cbl
                else:
                    # no logo found, just print the artist name instead
                    current_artist_label = InfoLargeLabel(text = result['artist'],font_size=Helpers.getfontsize(result['artist']))
                # clear the background of the player canvas so we can show the artist background
                self.ids.player.canvas.before.clear()
                # pick an artistid at random for the background
                mb_aid=random.choice(artistids)
                try:
                    # look in the 'artistbackground' subfolder of the artistid path
                    ab_path=os.path.join(fa_path,mb_aid,'artistbackground')
                    # pick one at random
                    img_path=random.choice(os.listdir(ab_path))
                    # update the player background with the image
                    with self.ids.player.canvas.before:
                        Color(1,1,1)
                        Rectangle(source=os.path.join(ab_path,img_path),size=self.ids.player.size,pos=self.ids.player.pos)
                except:
                    # if we can't get an artistbackground image, just do nothing
                    pass
                if os.path.isfile(p):
                    Logger.debug('update_mpd_currentsong: found good file at path '+p)
                    # load up the file to read the tags
                    f = mutagen.File(p)
                    cimg = None
                    data = None
                    originalyear = None
                    # if config file says use originalyear, use it instead of mpd's year
                    if self.config.getboolean('system','originalyear'):
                        if 'TXXX:originalyear' in f.keys():
                            originalyear=format(f['TXXX:originalyear'])
                    # try to get mp3 cover, if this throws an exception it's not an mp3 or it doesn't have a cover
                    try:
                        pframes = f.tags.getall("APIC")
                        # id3v2 can store any number of image frames, we just want the first one
                        for frame in pframes:
                            ext = 'img'
                            # figure out the file type
                            if frame.mime.endswith('jpeg') or frame.mime.endswith('jpg'):
                                ext = 'jpg'
                            elif frame.mime.endswith('png'):
                                ext = 'png'
                            elif frame.mime.endswith('bmp'):
                                ext = 'bmp'
                            elif frame.mime.endswith('gif'):
                                ext = 'gif'
                            # pull the raw image data into a variable
                            data=io.BytesIO(bytearray(frame.data))
                            break
                    except AttributeError:
                        pass
                    # try to get mp4 cover
                    if 'covr' in f.keys():
                        # figure out the file type
                        if f['covr'][0].imageformat == mutagen.mp4.MP4Cover.FORMAT_PNG:
                            ext = 'png'
                        else:
                            ext = 'jpg'
                        # pull the raw image data into a variable
                        data=io.BytesIO(bytearray(f['covr'][0]))
                    if data:
                        # if we got image data, load it as a kivy.core.image
                        # filter through PIL first to resize it if it is too large for a texture
                        pimg = PImage.open(data)
                        (w,h) = pimg.size
                        if w > max_cover_size or h > max_cover_size:
                            Logger.debug('update_mpd_currentsong: resizing cover image to maximum of '+format(max_cover_size)+'x'+format(max_cover_size))
                            pimg.thumbnail((max_cover_size,max_cover_size))
                            data2=io.BytesIO()
                            pimg.save(data2,'PNG')
                            data2.seek(0)
                            cimg = CoreImage(data2,ext='png')
                        else:
                            Logger.debug('update_mpd_currentsong: pulling cover directly from tag')
                            data.seek(0)
                            cimg = CoreImage(data,ext=ext)
                    if self.config.getboolean('system','originalyear') and originalyear and year and int(originalyear)!=int(year):
                        self.ids.yearlabel.text="["+originalyear+"]"
                        self.ids.remasterlabel.text="{"+year+"}"
                    elif year:
                        self.ids.yearlabel.text="["+year+"]"
                        self.ids.remasterlabel.text=""
                    else:
                        self.ids.yearlabel.text=""
                        self.ids.remasterlabel.text=""
                    if cimg:
                        img=CoverButton(img=cimg,layout=self.ids.album_cover_layout,halign='center')
                    else:
                        img=CoverButton(img=CoreImage(clearimage),layout=self.ids.album_cover_layout,halign='center')
                    self.ids.album_cover_layout.add_widget(img)
                    # popup the cover large if you press it
                    img.bind(on_press=partial(self.cover_popup,originalyear,year,result['album'],result['albumartist']))
                else:
                    # this should _probably_ never happen
                    Logger.debug('NowPlaying: no file found at path '+p)
                # add the correct artist name widget
                ti.add_widget(current_artist_label)
                lyt = BoxLayout(orientation='vertical',padding_y='2sp')
                if self.config.getboolean('system','advancedtitles'):
                    # check to see if song title has any data deliminated by () or []
                    stitle=re.split('[\(\[\]\)]',result['title'])
                    # if the first item is empty, the title starts with () or []
                    if not stitle[0]:
                        stitle.pop(0)
                        stitle[0]='('+stitle[0]+')'
                    ftitle=filter(None,stitle)
                    Logger.debug('TITLE: '+format(stitle))
                    if len(ftitle) > 1:
                        lyt2=BoxLayout(orientation='vertical',padding_y='2sp')
                        # split the title up and put the parentheses in smaller text below
                        lyt2.add_widget(InfoLargeLabel(text = stitle[0], font_size=Helpers.getfontsize(stitle[0])))
                        l2=""
                        for i, v in enumerate(stitle):
                            if i > 0 and v.strip():
                                if l2: l2=l2+' '
                                l2=l2+'('+v.strip()+')'
                        lyt2.add_widget(InfoLargeLabel(text = l2, font_size=Helpers.getfontsize(l2,2)))
                        lyt.add_widget(lyt2)
                    else:
                        lyt.add_widget(InfoLargeLabel(text = result['title'],font_size=Helpers.getfontsize(result['title'])))
                    # check to see if album is a single or EP
                    amatch=re.match(r'(.*) (\(single\)|EP)( .*)',result['album'])
                    # check if ' EP' is at the very end of the album title
                    if not amatch: amatch=re.match(r'(.*) (\(single\)|EP)($)',result['album'])
                    if amatch:
                        special=str(amatch.group(2)).strip("()")
                        Logger.debug("ALBUM: special release: "+special)
                        self.ids.releasetypelabel.text=special
                        galbum=str(amatch.group(1))+str(amatch.group(3))
                    else:
                        galbum=result['album']
                        self.ids.releasetypelabel.text='album'
                    # check to see if album is an import
                    amatch=re.match(r'(.*) (\(.. Import\))(.*)',galbum)
                    if amatch:
                        aimport=str(amatch.group(2))
                        Logger.debug("ALBUM: import: "+aimport)
                        galbum=str(amatch.group(1))+str(amatch.group(3))
                    else:
                        aimport=None
                    # check to see if album title is a split (has a ' / ' in the middle)
                    talbum=galbum.split(' / ')
                    Logger.debug('ALBUM1: '+format(talbum))
                    lyt2=BoxLayout(orientation='horizontal',padding_x='2sp')
                    if len(talbum)>1:
                        Logger.debug('ALBUM1: album is a split')
                        split=True
                    else:
                        split=False
                    for j, a in enumerate(talbum):
                        if j>0: lyt2.add_widget(InfoLargeLabel(font_size='50sp',text=u'\u2571',size_hint_x=None,font_name=normalfont))
                        # check to see if album title has any data deliminated by () or []
                        salbum=re.split('[\(\[\]\)]',a)
                        Logger.debug('ALBUM2: '+format(salbum))
                        if len(salbum) > 1:
                            lyt3=BoxLayout(orientation='vertical',padding_y='2sp')
                            # split the album up and put the parentheses in smaller text below
                            if split:
                                lyt3.add_widget(InfoLargeLabel(text = salbum[0], font_size=Helpers.getfontsize(salbum[0],1.5)))
                            else:
                                lyt3.add_widget(InfoLargeLabel(text = salbum[0], font_size=Helpers.getfontsize(salbum[0])))
                            l2=""
                            for i, v in enumerate(salbum):
                                if i > 0 and v.strip():
                                    if l2: l2=l2+' '
                                    l2=l2+'('+v.strip()+')'
                            lyt3.add_widget(InfoLargeLabel(text = l2, font_size=Helpers.getfontsize(l2,2)))
                            lyt2.add_widget(lyt3)
                        else:
                            if split:
                                lyt2.add_widget(InfoLargeLabel(text = a, font_size=Helpers.getfontsize(a,1.5)))
                            else:
                                lyt2.add_widget(InfoLargeLabel(text = a, font_size=Helpers.getfontsize(a)))
                    lyt.add_widget(lyt2)
                    if aimport: lyt.add_widget(InfoLargeLabel(text=aimport,font_size=Helpers.getfontsize(aimport,2)))
                else:
                    self.ids.releasetypelabel.text=''
                    lyt.add_widget(InfoLargeLabel(text = result['title'],font_size=Helpers.getfontsize(result['title'])))
                    lyt.add_widget(InfoLargeLabel(text = result['album'],font_size=Helpers.getfontsize(result['album'])))
                ti.add_widget(lyt)
        else:
            # there's not a current song, so zero everything out
            self.stop_zero_stuff()

    def update_mpd_sticker_rating(self,result):
        """Callback for song that has a rating in mpd."""
        Logger.debug('NowPlaying: update_mpd_sticker_rating')
        self.ids.song_star_button.disabled=False
        self.ids.song_star_button.text=ratingstars[int(result)]

    def handle_mpd_no_sticker(self,result):
        """Callback for song that has no rating in mpd."""
        Logger.debug('NowPlaying: handle_mpd_no_sticker')
        self.ids.song_star_button.disabled=False
        self.ids.song_star_button.text=ratingstars[11]

    def update_mpd_nextsong(self,result):
        """Callback for next song data from mpd."""
        Logger.debug('NowPlaying: update_mpd_nextsong()')
        # result is a list with one member
        for obj in result:
            # set the next song label
            self.ids.next_song_artist_label.text = 'Up Next: '+obj['artist']+' - '+obj['title']

    def prev_pressed(self):
        """Callback for prev button pressed."""
        Logger.debug('Application: prev_pressed()')
        self.mpdconnection.protocol.previous()

    def play_pressed(self):
        """Callback for play/pause button pressed."""
        Logger.debug('Application: play_pressed()')
        if self.mpd_status['state'] == 'play':
            # pause if playing
            self.mpdconnection.protocol.pause()
        else:
            # play if paused or stopped
            self.mpdconnection.protocol.play()

    def next_pressed(self):
        """Callback for next button pressed."""
        Logger.debug('Application: next_pressed()')
        self.mpdconnection.protocol.next()

    def repeat_pressed(self):
        """Callback for repeat button pressed."""
        Logger.debug('Application: repeat_pressed()')
        # toggle on/off
        self.mpdconnection.protocol.repeat(str(1-int(self.mpd_status['repeat'])))

    def single_pressed(self):
        """Callback for single button pressed."""
        Logger.debug('Application: single_pressed()')
        # toggle on/off
        self.mpdconnection.protocol.single(str(1-int(self.mpd_status['single'])))

    def random_pressed(self):
        """Callback for random button pressed."""
        Logger.debug('Application: random_pressed()')
        # toggle on/off
        self.mpdconnection.protocol.random(str(1-int(self.mpd_status['random'])))

    def consume_pressed(self):
        """Callback for consume button pressed."""
        Logger.debug('Application: consume_pressed()')
        # toggle on/off
        self.mpdconnection.protocol.consume(str(1-int(self.mpd_status['consume'])))

    def rating_popup(self,instance):
        """Popup for setting song rating."""
        Logger.debug('Application: rating_popup()')
        popup=Factory.RatingPopup(rating_set=self.rating_set,song=self.currfile)
        popup.open()

    def rating_set(self,song,rating,popup):
        """Method that sets a song's rating."""
        Logger.debug('Application: rating_set('+rating+')')
        popup.dismiss()
        if rating:
            self.mpdconnection.protocol.sticker_set('song',song,'rating',rating)
        else:
            self.mpdconnection.protocol.sticker_delete('song',song,'rating')

    def cover_popup(self,originalyear,year,album,albumartist,instance):
        """Popup for showing a larger version of the album cover."""
        Logger.debug('Application: cover_popup()')
        title='"'+album+'"'
        if year and originalyear and year!=originalyear:
            title=title+' released '+originalyear+' (remastered '+year+')'
        elif year:
            title=title+' released '+year
        title=title+' by '+albumartist
        popup=Factory.CoverPopup(title=title,image_texture=instance.img.texture)
        popup.open()

    def change_backlight(self,value):
        """Method that sets the backlight to a certain value."""
        Logger.info('Application: change_backlight('+str(value)+') rpienable = '+self.config.get('system','rpienable'))
        # only if the ini file says it's ok
        if self.config.getboolean('system','rpienable'):
            import rpi_backlight as bl
            # set the brightness
            bl.set_brightness(int(value), smooth=True, duration=1)

    def settings_update(self):
        self.mpdconnection.protocol.currentsong().addCallback(partial(self.update_mpd_currentsong,True)).addErrback(self.mpdconnection.handle_mpd_error)

    def init_mpd(self,instance):
        # get the initial mpd status
        self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.mpdconnection.handle_mpd_error)
        # create the once-per-second update of the track slider
        self.track_slider_task=Clock.schedule_interval(self.update_track_slider,1)
        self.track_slider_task.cancel()
        # subscribe to 'kmpc' to check for messages from mpd
        self.mpdconnection.protocol.subscribe('kmpc')

    def mpd_idle_handler(self,result):
        # global flag for disabling during sync
        if self.do_idle_handler:
            # notify various subsystems based on what changed
            for s in result:
                Logger.info('mpd_idle_handler: Changed '+format(s))
                if format(s) == 'playlist':
                    # playlist was changed, ask mpd for playlist info
                    self.mpdconnection.protocol.playlistinfo().addCallback(self.ids.playlist_tab.populate_playlist).addErrback(self.ids.playlist_tab.handle_mpd_error)
                    # force a reload of nextsong if playlist changes
                    self.nextsong = None
                    self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
                elif format(s) == 'player':
                    # player was changed, ask mpd for player status
                    self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
                elif format(s) == 'sticker':
                    # song rating sticker was changed, ask mpd for current song rating
                    self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
                    self.mpdconnection.protocol.sticker_get('song',self.currfile,'rating').addCallback(self.update_mpd_sticker_rating).addErrback(self.handle_mpd_no_sticker)
                elif format(s) == 'options':
                    # some playback option was changed, ask mpd for player status
                    self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
                elif format(s) == 'message':
                    # an mpd message was received, ask mpd what it was
                    self.mpdconnection.protocol.readmessages().addCallback(self.handle_mpd_message).addErrback(self.handle_mpd_error)
                else:
                    # default if none of the above, ask mpd for player status
                    self.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)



# python-mpd2: Python MPD client library
#
# Copyright (C) 2008-2010  J. Alexander Treuman <jat@spatialrift.net>
# Copyright (C) 2012  J. Thalheim <jthalheim@gmail.com>
# Copyright (C) 2016  Robert Niederreiter <rnix@squarewave.at>
#
# python-mpd2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-mpd2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-mpd2.  If not, see <http://www.gnu.org/licenses/>.

from collections import Callable
import logging
import socket
import sys
import warnings


###############################################################################
# constants
###############################################################################

VERSION = (0, 6, 1)
HELLO_PREFIX = "OK MPD "
ERROR_PREFIX = "ACK "
SUCCESS = "OK"
NEXT = "list_OK"


###############################################################################
# utils
###############################################################################

IS_PYTHON2 = sys.version_info < (3, 0)
if IS_PYTHON2:
    def decode_str(s):
        return s.decode("utf-8")

    def encode_str(s):
        if type(s) == str:
            return s
        else:
            return (unicode(s)).encode("utf-8")
else:
    def decode_str(s):
        return s

    encode_str = str


def escape(text):
    return text.replace("\\", "\\\\").replace('"', '\\"')


###############################################################################
# logging
###############################################################################

try:
    from logging import NullHandler
except ImportError:  # NullHandler was introduced in python2.7
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())


###############################################################################
# exceptions
###############################################################################

class MPDError(Exception):
    pass


class ConnectionError(MPDError):
    pass


class ProtocolError(MPDError):
    pass


class CommandError(MPDError):
    pass


class CommandListError(MPDError):
    pass


class PendingCommandError(MPDError):
    pass


class IteratingError(MPDError):
    pass


###############################################################################
# command registration
###############################################################################

class mpd_commands(object):
    """Decorator for registering MPD commands with it's corresponding result
    callback.
    """

    def __init__(self, *commands):
        self.commands = commands

    def __call__(self, ob):
        ob.mpd_commands = self.commands
        return ob


def mpd_command_provider(cls):
    """Decorator hooking up registered MPD commands to concrete client
    implementation.

    A class using this decorator must inherit from ``MPDClientBase`` and
    implement it's ``add_command`` function.
    """
    def collect(cls, callbacks=dict()):
        """Collect MPD command callbacks from given class.

        Searches class __dict__ on given class and all it's bases for functions
        which have been decorated with @mpd_commands and returns a dict
        containing callback name as keys and
        (callback, callback implementing class) tuples as values.
        """
        for name, ob in cls.__dict__.items():
            if hasattr(ob, "mpd_commands") and name not in callbacks:
                callbacks[name] = (ob, cls)
        for base in cls.__bases__:
            callbacks = collect(base, callbacks)
        return callbacks

    for name, value in collect(cls).items():
        callback, from_ = value
        for command in callback.mpd_commands:
            cls.add_command(command, callback)
    return cls


class Noop(object):
    """An instance of this class represents a MPD command callback which
    does nothing.
    """
    mpd_commands = None


###############################################################################
# abstract base client
###############################################################################

class MPDClientBase(object):
    """Abstract MPD client.

    This class defines a general client contract, provides MPD protocol parsers
    and defines all available MPD commands and it's corresponding result
    parsing callbacks. There might be the need of overriding some callbacks on
    subclasses.
    """

    def __init__(self, use_unicode=False):
        self.iterate = False
        self.use_unicode = use_unicode
        self._reset()

    @classmethod
    def add_command(cls, name, callback):
        raise NotImplementedError(
            'Abstract ``MPDClientBase`` does not implement ``add_command``')

    def noidle(self):
        raise NotImplementedError(
            'Abstract ``MPDClientBase`` does not implement ``noidle``')

    def command_list_ok_begin(self):
        raise NotImplementedError(
            'Abstract ``MPDClientBase`` does not implement '
            '``command_list_ok_begin``')

    def command_list_end(self):
        raise NotImplementedError(
            'Abstract ``MPDClientBase`` does not implement '
            '``command_list_end``')

    def _reset(self):
        self.mpd_version = None
        self._command_list = None

    def _parse_pair(self, line, separator):
        if line is None:
            return
        pair = line.split(separator, 1)
        if len(pair) < 2:
            raise ProtocolError("Could not parse pair: '{}'".format(line))
        return pair

    def _parse_pairs(self, lines, separator=": "):
        for line in lines:
            yield self._parse_pair(line, separator)

    def _parse_objects(self, lines, delimiters=[]):
        obj = {}
        for key, value in self._parse_pairs(lines):
            key = key.lower()
            if obj:
                if key in delimiters:
                    yield obj
                    obj = {}
                elif key in obj:
                    if not isinstance(obj[key], list):
                        obj[key] = [obj[key], value]
                    else:
                        obj[key].append(value)
                    continue
            obj[key] = value
        if obj:
            yield obj

    def _parse_raw_stickers(self, lines):
        for key, sticker in self._parse_pairs(lines):
            value = sticker.split('=', 1)
            if len(value) < 2:
                raise ProtocolError(
                    "Could not parse sticker: {}".format(repr(sticker)))
            yield tuple(value)

    NOOP = mpd_commands('close', 'kill')(Noop())

    @mpd_commands('plchangesposid')
    def _parse_changes(self, lines):
        return self._parse_objects(lines, ["cpos"])

    @mpd_commands('listall', 'listallinfo', 'listfiles', 'lsinfo')
    def _parse_database(self, lines):
        return self._parse_objects(lines, ["file", "directory", "playlist"])

    @mpd_commands('idle')
    def _parse_idle(self, lines):
        return self._parse_list(lines)

    @mpd_commands('addid', 'config', 'replay_gain_status', 'rescan', 'update')
    def _parse_item(self, lines):
        pairs = list(self._parse_pairs(lines))
        if len(pairs) != 1:
            return
        return pairs[0][1]

    @mpd_commands(
        'channels', 'commands', 'list', 'listplaylist', 'notcommands',
        'tagtypes', 'urlhandlers')
    def _parse_list(self, lines):
        seen = None
        for key, value in self._parse_pairs(lines):
            if key != seen:
                if seen is not None:
                    raise ProtocolError(
                        "Expected key '{}', got '{}'".format(seen, key))
                seen = key
            yield value

    @mpd_commands('readmessages')
    def _parse_messages(self, lines):
        return self._parse_objects(lines, ["channel"])

    @mpd_commands('listmounts')
    def _parse_mounts(self, lines):
        return self._parse_objects(lines, ["mount"])

    @mpd_commands('listneighbors')
    def _parse_neighbors(self, lines):
        return self._parse_objects(lines, ["neighbor"])

    @mpd_commands(
        'add', 'addtagid', 'clear', 'clearerror', 'cleartagid', 'consume',
        'crossfade', 'delete', 'deleteid', 'disableoutput', 'enableoutput',
        'findadd', 'load', 'mixrampdb', 'mixrampdelay', 'mount', 'move',
        'moveid', 'next', 'password', 'pause', 'ping', 'play', 'playid',
        'playlistadd', 'playlistclear', 'playlistdelete', 'playlistmove',
        'previous', 'prio', 'prioid', 'random', 'rangeid', 'rename', 'repeat',
        'replay_gain_mode', 'rm', 'save', 'searchadd', 'searchaddpl', 'seek',
        'seekcur', 'seekid', 'sendmessage', 'setvol', 'shuffle', 'single',
        'sticker delete', 'sticker set', 'stop', 'subscribe', 'swap', 'swapid',
        'toggleoutput', 'umount', 'unsubscribe')
    def _parse_nothing(self, lines):
        for line in lines:
            raise ProtocolError(
                "Got unexpected return value: '{}'".format(', '.join(lines)))

    @mpd_commands('count', 'currentsong', 'readcomments', 'stats', 'status')
    def _parse_object(self, lines):
        objs = list(self._parse_objects(lines))
        if not objs:
            return {}
        return objs[0]

    @mpd_commands('outputs')
    def _parse_outputs(self, lines):
        return self._parse_objects(lines, ["outputid"])

    @mpd_commands('playlist')
    def _parse_playlist(self, lines):
        for key, value in self._parse_pairs(lines, ":"):
            yield value

    @mpd_commands('listplaylists')
    def _parse_playlists(self, lines):
        return self._parse_objects(lines, ["playlist"])

    @mpd_commands('decoders')
    def _parse_plugins(self, lines):
        return self._parse_objects(lines, ["plugin"])

    @mpd_commands(
        'find', 'listplaylistinfo', 'playlistfind', 'playlistid',
        'playlistinfo', 'playlistsearch', 'plchanges', 'search', 'sticker find')
    def _parse_songs(self, lines):
        return self._parse_objects(lines, ["file"])

    @mpd_commands('sticker get')
    def _parse_sticker(self, lines):
        key, value = list(self._parse_raw_stickers(lines))[0]
        return value

    @mpd_commands('sticker list')
    def _parse_stickers(self, lines):
        return dict(self._parse_raw_stickers(lines))


###############################################################################
# sync client
###############################################################################

def _create_callback(self, function, wrap_result):
    """Create MPD command related response callback.
    """
    if not isinstance(function, Callable):
        return None
    def command_callback():
        # command result callback expects response from MPD as iterable lines,
        # thus read available lines from socket
        res = function(self, self._read_lines())
        # wrap result in iterator helper if desired
        if wrap_result:
            res = self._wrap_iterator(res)
        return res
    return command_callback


def _create_command(wrapper, name, return_value, wrap_result):
    """Create MPD command related function.
    """
    def mpd_command(self, *args):
        callback = _create_callback(self, return_value, wrap_result)
        return wrapper(self, name, args, callback)
    return mpd_command


class _NotConnected(object):
    def __getattr__(self, attr):
        return self._dummy

    def _dummy(*args):
        raise ConnectionError("Not connected")


@mpd_command_provider
class MPDClient(MPDClientBase):
    idletimeout = None
    _timeout = None
    _wrap_iterator_parsers = [
        MPDClientBase._parse_list,
        MPDClientBase._parse_playlist,
        MPDClientBase._parse_changes,
        MPDClientBase._parse_songs,
        MPDClientBase._parse_mounts,
        MPDClientBase._parse_neighbors,
        MPDClientBase._parse_playlists,
        MPDClientBase._parse_database,
        MPDClientBase._parse_messages,
        MPDClientBase._parse_outputs,
        MPDClientBase._parse_plugins
    ]
    if IS_PYTHON2:
        _wrap_iterator_parsers = [f.__func__ for f in _wrap_iterator_parsers]

    def __init__(self, use_unicode=False):
        super(MPDClient, self).__init__(use_unicode=use_unicode)

    def _reset(self):
        super(MPDClient, self)._reset()
        self._pending = []
        self._iterating = False
        self._sock = None
        self._rfile = _NotConnected()
        self._wfile = _NotConnected()

    def _send(self, command, args, retval):
        warnings.warn(
            "``send_{}`` is deprecated in favor of "
            "asynchronous API".format(command),
            DeprecationWarning)
        if self._command_list is not None:
            raise CommandListError(
                "Cannot use send_{} in a command list".format(command))
        self._write_command(command, args)
        if retval is not None:
            self._pending.append(command)

    def _fetch(self, command, args, retval):
        warnings.warn(
            "``fetch_{}`` is deprecated in favor of "
            "asynchronous API".format(command),
            DeprecationWarning)
        if self._command_list is not None:
            raise CommandListError(
                "Cannot use fetch_{} in a command list".format(command))
        if self._iterating:
            raise IteratingError(
                "Cannot use fetch_{} while iterating".format(command))
        if not self._pending:
            raise PendingCommandError("No pending commands to fetch")
        if self._pending[0] != command:
            raise PendingCommandError(
                "'{}' is not the currently pending command".format(command))
        del self._pending[0]
        if isinstance(retval, Callable):
            return retval()
        return retval

    def _execute(self, command, args, retval):
        if self._iterating:
            raise IteratingError(
                "Cannot execute '{}' while iterating".format(command))
        if self._pending:
            raise PendingCommandError(
                "Cannot execute '{}' with pending commands".format(command))
        if self._command_list is not None:
            if not isinstance(retval, Callable):
                raise CommandListError(
                    "'{}' not allowed in command list".format(command))
            self._write_command(command, args)
            self._command_list.append(retval)
        else:
            self._write_command(command, args)
            if isinstance(retval, Callable):
                return retval()
            return retval

    def _write_line(self, line):
        self._wfile.write("{}\n".format(line))
        self._wfile.flush()

    def _write_command(self, command, args=[]):
        parts = [command]
        for arg in args:
            if type(arg) is tuple:
                if len(arg) == 0:
                    parts.append('":"')
                elif len(arg) == 1:
                    parts.append('"{}:"'.format(int(arg[0])))
                else:
                    parts.append('"{}:{}"'.format(int(arg[0]), int(arg[1])))
            else:
                parts.append('"{}"'.format(escape(encode_str(arg))))
        # Minimize logging cost if the logging is not activated.
        if logger.isEnabledFor(logging.DEBUG):
            if command == "password":
                logger.debug("Calling MPD password(******)")
            else:
                logger.debug("Calling MPD %s%r", command, args)
        cmd = " ".join(parts)
        self._write_line(cmd)

    def _read_line(self):
        line = self._rfile.readline()
        if self.use_unicode:
            line = decode_str(line)
        if not line.endswith("\n"):
            self.disconnect()
            raise ConnectionError("Connection lost while reading line")
        line = line.rstrip("\n")
        if line.startswith(ERROR_PREFIX):
            error = line[len(ERROR_PREFIX):].strip()
            raise CommandError(error)
        if self._command_list is not None:
            if line == NEXT:
                return
            if line == SUCCESS:
                raise ProtocolError("Got unexpected '{}'".format(SUCCESS))
        elif line == SUCCESS:
            return
        return line

    def _read_lines(self):
        line = self._read_line()
        while line is not None:
            yield line
            line = self._read_line()

    def _read_command_list(self):
        try:
            for retval in self._command_list:
                yield retval()
        finally:
            self._command_list = None
        self._parse_nothing(self._read_lines())

    def _iterator_wrapper(self, iterator):
        try:
            for item in iterator:
                yield item
        finally:
            self._iterating = False

    def _wrap_iterator(self, iterator):
        if not self.iterate:
            return list(iterator)
        self._iterating = True
        return self._iterator_wrapper(iterator)

    def _hello(self):
        line = self._rfile.readline()
        if not line.endswith("\n"):
            self.disconnect()
            raise ConnectionError("Connection lost while reading MPD hello")
        line = line.rstrip("\n")
        if not line.startswith(HELLO_PREFIX):
            raise ProtocolError("Got invalid MPD hello: '{}'".format(line))
        self.mpd_version = line[len(HELLO_PREFIX):].strip()

    def _connect_unix(self, path):
        if not hasattr(socket, "AF_UNIX"):
            raise ConnectionError(
                "Unix domain sockets not supported on this platform")
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect(path)
        return sock

    def _connect_tcp(self, host, port):
        try:
            flags = socket.AI_ADDRCONFIG
        except AttributeError:
            flags = 0
        err = None
        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC,
                                      socket.SOCK_STREAM, socket.IPPROTO_TCP,
                                      flags):
            af, socktype, proto, canonname, sa = res
            sock = None
            try:
                sock = socket.socket(af, socktype, proto)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                sock.settimeout(self.timeout)
                sock.connect(sa)
                return sock
            except socket.error as e:
                err = e
                if sock is not None:
                    sock.close()
        if err is not None:
            raise err
        else:
            raise ConnectionError("getaddrinfo returns an empty list")

    @mpd_commands('idle')
    def _parse_idle(self, lines):
        self._sock.settimeout(self.idletimeout)
        ret = self._wrap_iterator(self._parse_list(lines))
        self._sock.settimeout(self._timeout)
        return ret

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout
        if self._sock is not None:
            self._sock.settimeout(timeout)

    def connect(self, host, port=None, timeout=None):
        logger.info(
            "Calling MPD connect(%r, %r, timeout=%r)", host, port, timeout)
        if self._sock is not None:
            raise ConnectionError("Already connected")
        if timeout is not None:
            warnings.warn(
                "The timeout parameter in connect() is deprecated! "
                "Use MPDClient.timeout = yourtimeout instead.",
                DeprecationWarning)
            self.timeout = timeout
        if host.startswith("/"):
            self._sock = self._connect_unix(host)
        else:
            if port is None:
                raise ValueError("port argument must be specified when connecting via tcp")
            self._sock = self._connect_tcp(host, port)
        if IS_PYTHON2:
            self._rfile = self._sock.makefile("r")
            self._wfile = self._sock.makefile("w")
        else:
            # - Force UTF-8 encoding, since this is dependant from the LC_CTYPE
            #   locale.
            # - by setting newline explicit, we force to send '\n' also on
            #   windows
            self._rfile = self._sock.makefile(
                "r",
                encoding="utf-8",
                newline="\n")
            self._wfile = self._sock.makefile(
                "w",
                encoding="utf-8",
                newline="\n")
        try:
            self._hello()
        except:
            self.disconnect()
            raise

    def disconnect(self):
        logger.info("Calling MPD disconnect()")
        if (self._rfile is not None and
                not isinstance(self._rfile, _NotConnected)):
            self._rfile.close()
        if (self._wfile is not None and
                not isinstance(self._wfile, _NotConnected)):
            self._wfile.close()
        if self._sock is not None:
            self._sock.close()
        self._reset()

    def fileno(self):
        if self._sock is None:
            raise ConnectionError("Not connected")
        return self._sock.fileno()

    def noidle(self):
        if not self._pending or self._pending[0] != 'idle':
            msg = 'cannot send noidle if send_idle was not called'
            raise CommandError(msg)
        del self._pending[0]
        self._write_command("noidle")
        return self._wrap_iterator(self._parse_list(self._read_lines()))

    def command_list_ok_begin(self):
        if self._command_list is not None:
            raise CommandListError("Already in command list")
        if self._iterating:
            raise IteratingError("Cannot begin command list while iterating")
        if self._pending:
            raise PendingCommandError(
                "Cannot begin command list with pending commands")
        self._write_command("command_list_ok_begin")
        self._command_list = []

    def command_list_end(self):
        if self._command_list is None:
            raise CommandListError("Not in command list")
        if self._iterating:
            raise IteratingError("Already iterating over a command list")
        self._write_command("command_list_end")
        return self._wrap_iterator(self._read_command_list())

    @classmethod
    def add_command(cls, name, callback):
        wrap_result = callback in cls._wrap_iterator_parsers
        method = _create_command(cls._execute, name, callback, wrap_result)
        send_method = _create_command(cls._send, name, callback, wrap_result)
        fetch_method = _create_command(cls._fetch, name, callback, wrap_result)
        # create new mpd commands as function in three flavors:
        # normal, with "send_"-prefix and with "fetch_"-prefix
        escaped_name = name.replace(" ", "_")
        setattr(cls, escaped_name, method)
        setattr(cls, "send_" + escaped_name, send_method)
        setattr(cls, "fetch_" + escaped_name, fetch_method)

    @classmethod
    def remove_command(cls, name):
        if not hasattr(cls, name):
            raise ValueError(
                "Can't remove not existent '{}' command".format(name))
        name = name.replace(" ", "_")
        delattr(cls, str(name))
        delattr(cls, str("send_" + name))
        delattr(cls, str("fetch_" + name))

# vim: set expandtab shiftwidth=4 softtabstop=4 textwidth=79:



from __future__ import print_function
import os
import io
from functools import partial
from kmpc.mpd import MPDProtocol
from kmpc.mpdfactory import MpdConnection
from kmpc.extra import KmpcHelpers

import kivy
kivy.require('1.10.0')

from twisted.internet import protocol
from twisted.internet.defer import Deferred,DeferredList

from kivy.logger import Logger

Helpers=KmpcHelpers()

class Sync(object):

    def __init__(self,config,runparts=[]):
        from twisted.internet import reactor
        self.config=config
        self.mpdhost=config.get('mpd','mpdhost')
        self.mpdport=config.get('mpd','mpdport')
        self.synchost=config.get('sync','synchost')
        self.syncmpdport=config.get('sync','syncmpdport')
        self.synclist=config.get('sync','syncplaylist')
        self.basepath=config.get('paths','musicpath')
        self.fanartpath=config.get('paths','fanartpath')
        self.syncbasepath=config.get('sync','syncmusicpath')
        self.syncfanartpath=config.get('sync','syncfanartpath')
        self.tmppath=config.get('paths','tmppath')
        self.runparts=runparts
        self.localconnected=False
        self.syncconnected=False
        self.kivy=True
        self.filelist={}
        self.callbacks=[]
        self.pltotal=0
        self.pldone=0
        self.print_line("Running sync with synchost "+self.synchost+" with modules "+format(runparts))
        if self.mpdhost==self.synchost:
            Logger.warn('Sync: will not sync identical hosts!')
            return
        self.localmpd=MpdConnection(self.config,self.mpdhost,self.mpdport,self.mpd_idle_handler,[self.init_local_mpd],True,False)
        self.syncmpd=MpdConnection(self.config,self.synchost,self.syncmpdport,None,[self.init_sync_mpd],True,False)
        self.d = Deferred()
        self.d2 = Deferred()
        self.d3 = Deferred()
        self.d4 = Deferred()
        if 'music' in runparts:
            self.callbacks.append(self.d.addCallbacks(self.sync_music,self.errback))
            self.callbacks.append(self.d.addCallbacks(self.build_filelist,self.errback))
            self.callbacks.append(self.d.addCallbacks(self.cleanup_music_sync,self.errback))
            self.callbacks.append(self.d2.addErrback(self.errback))
        if 'fanart' in runparts:
            self.callbacks.append(self.d.addCallbacks(self.sync_fanart,self.errback))
            self.callbacks.append(self.d.addCallbacks(self.finish_fanart_sync,self.errback))
        if 'exportratings' in runparts:
            self.callbacks.append(self.d.addCallbacks(self.sync_export_ratings,self.errback))
            self.callbacks.append(self.d.addCallbacks(self.handle_export_ratings,self.errback))
            self.callbacks.append(self.d3.addErrback(self.errback))
        if 'importratings' in runparts:
            self.callbacks.append(self.d.addCallbacks(self.sync_import_ratings,self.errback))
            self.callbacks.append(self.d.addCallbacks(self.handle_import_ratings,self.errback))
            self.callbacks.append(self.d4.addErrback(self.errback))
        if not reactor.running:
            self.kivy=False
            reactor.run()

#### override these when subclassing
    def run_at_end(self,result):
        from twisted.internet import reactor
        Logger.debug("Sync: callbacks done: "+format(result))
        if not self.kivy: reactor.stop()

    def show_ratings_progress(self,done,total,sdir):
        print(done,'of',total,end='\r')

    def ratings_incoming(self,sdir):
        pass

    def print_line(self,line):
        Logger.info("Sync: "+line)

    def errback(self,result):
        Logger.error('Sync: Callback error: {}'.format(result))

#### overall operation functions
    def init_local_mpd(self,conn):
        Logger.debug("init_local_mpd: connected")
        self.localconnected=True
        self.do_sync()

    def init_sync_mpd(self,conn):
        Logger.debug("init_sync_mpd: connected")
        self.syncconnected=True
        self.do_sync()

    def do_sync(self):
        if self.localconnected and self.syncconnected:
            self.print_line("Beginning sync process")
            dlist=DeferredList(self.callbacks,consumeErrors=True)
            dlist.addCallbacks(self.disconnect,self.errback)
            dlist.addCallbacks(self.run_at_end,self.errback)
            self.d.callback('BEGIN')

    def disconnect(self,result):
        self.print_line("Disconnecting from synchost")
        try:
            self.localmpd.reactor.disconnect()
            self.syncmpd.reactor.disconnect()
        except Exception as e:
            Logger.debug("Exception: "+format(e))
        return True

#### fanart sync functions
    def sync_fanart(self,result):
        from twisted.internet import reactor
        self.print_line("Syncing fanart")
        # rsync the files
        cmdline=[
            'rsync',
            '-vruxhm',
            self.synchost+':'+self.syncfanartpath+'/',
            self.fanartpath
            ]
        pp=Subproc(self.print_line)
        pp.deferred=Deferred()
        p=reactor.spawnProcess(pp,cmdline[0],cmdline,{})
        return pp.deferred

    def finish_fanart_sync(self,result):
        self.print_line("Fanart synced from synchost")
        return True

#### ratings sync functions
    def sync_export_ratings(self,result):
        self.print_line("Exporting ratings")
        return self.localmpd.protocol.sticker_find('song','','rating')

    def handle_export_ratings(self,result):
        Logger.debug("Sync: handle_export_ratings")
        cb=[]
        i=1
        rlist=list(result)
        total=len(rlist)
        self.ratings_incoming('Export')
        for row in rlist:
            uri=Helpers.decodeFileName(row['file'])
            rating=str(row['sticker'].split('=')[1])
            cb.append(self.syncmpd.protocol.sticker_set('song',uri,'rating',rating).addBoth(partial(self.handle_rating_set,'Export',uri,rating,i,total)))
            i=i+1
        udl=DeferredList(cb,consumeErrors=True)
        return udl.addCallbacks(self.finish_export_ratings,self.errback)

    def finish_export_ratings(self,result):
        Logger.debug("Sync: finish_export_ratings")
        self.d3.callback(True)
        return True

    def sync_import_ratings(self,result):
        self.print_line("Importing ratings")
        return self.syncmpd.protocol.sticker_find('song','','rating')

    def handle_import_ratings(self,result):
        Logger.debug("Sync: handle_import_ratings")
        cb=[]
        i=1
        rlist=list(result)
        total=len(rlist)
        self.ratings_incoming('Import')
        for row in rlist:
            uri=Helpers.decodeFileName(row['file'])
            rating=str(row['sticker'].split('=')[1])
            cb.append(self.localmpd.protocol.sticker_set('song',uri,'rating',rating).addBoth(partial(self.handle_rating_set,'Import',uri,rating,i,total)))
            i=i+1
        udl=DeferredList(cb,consumeErrors=True)
        return udl.addCallbacks(self.finish_import_ratings,self.errback)

    def finish_import_ratings(self,result):
        Logger.debug("Sync: finish_import_ratings")
        self.d4.callback(True)
        return True

    def handle_rating_set(self,sdir,uri,rating,done,total,result):
        self.show_ratings_progress(done,total,sdir)

#### music sync functions
    def sync_music(self,result):
        self.print_line("Syncing music using playlist ["+self.synclist+"]")
        # clear the local 'root' playlist
        self.localmpd.protocol.playlistclear('root').addErrback(self.localmpd.handle_mpd_error)
        return self.syncmpd.protocol.listplaylist(self.synclist)

    def build_filelist(self,result):
        from twisted.internet import reactor
        Logger.debug("build_filelist: writing file/hash")
        # write synclist to a temp file, and a hash
        f=io.open(os.path.join(self.tmppath,'rsync.inc'),mode='w',encoding="utf-8")
        for row in result:
            f.write(Helpers.decodeFileName(row)+"\n")
            self.filelist[Helpers.decodeFileName(row.rstrip())]=True
        f.close()
        # this whole block walks the filesystem and deletes any file that is not in the synclist
        Logger.debug("build_filelist: deleting old files")
        for dirpath, dirnames, filenames in os.walk(Helpers.decodeFileName(self.basepath)):
            if len(filenames)>0:
                rpath = dirpath[len(self.basepath+os.sep):]
                for filename in filenames:
                    fpath=os.path.join(Helpers.decodeFileName(rpath),Helpers.decodeFileName(filename))
                    apath = os.path.join(Helpers.decodeFileName(dirpath),Helpers.decodeFileName(filename))
                    if fpath not in self.filelist:
                        Logger.debug("Filesync: Deleting "+apath)
                        os.remove(Helpers.decodeFileName(apath))
        Logger.debug("build_filelist: removing empty folders")
        # remove empty folders
        Helpers.removeEmptyFolders(self.basepath)
        # rsync the files
        cmdline=[
            'rsync',
            '-vruxhm',
            '--files-from='+os.path.join(self.tmppath,'rsync.inc'),
            self.synchost+':'+self.syncbasepath+'/',
            self.basepath
            ]
        pp=Subproc(self.print_line)
        pp.deferred=Deferred()
        p=reactor.spawnProcess(pp,cmdline[0],cmdline,{})
        return pp.deferred

    def cleanup_music_sync(self,result):
        self.print_line("Cleaning up temp files and updating mpd database")
        # clean up
        os.remove(os.path.join(self.tmppath,'rsync.inc'))
        self.localmpd.protocol.update()
        return True

    def mpd_idle_handler(self,result):
        for s in result:
            if format(s)=='update':
                Logger.debug('mpd_idle_handler: Changed '+format(s))
                self.localmpd.protocol.status().addCallbacks(self.handle_update,self.errback)

    def handle_update(self,result):
        if 'updating_db' not in result:
            Logger.debug("handle_update: updating_db done")
            cb=[]
            for k in sorted(self.filelist.keys()):
                cb.append(self.localmpd.protocol.playlistadd('root',k))
            udl=DeferredList(cb,consumeErrors=True)
            udl.addCallbacks(self.finish_update,self.errback)
        return True

    def finish_update(self,result):
        self.print_line("Music synced and playlist updated")
        self.d2.callback(True)
        return True

#### class to handle subprocesses

class Subproc(protocol.ProcessProtocol):

    def __init__(self,printer):
        self.printer=printer

    def connectionMade(self):
        Logger.debug("Subproc: spawning subprocess")

    def processExited(self,reason):
        Logger.debug("Subproc: exited, status: "+format(reason.value.exitCode))
        self.deferred.callback(True)

    def outReceived(self,data):
        for line in data.split('\n'):
            if line: self.printer(line)



import os
from pkg_resources import resource_filename

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

# import all the other kivy stuff
from kivy.app import App
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.slider import Slider
from kivy.uix.label import Label
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.graphics import Rectangle
from kivy.uix.image import Image,AsyncImage
from kivy.uix.behaviors import ButtonBehavior
from kivy.properties import ObjectProperty,StringProperty,NumericProperty
from kivy.uix.behaviors import ButtonBehavior, FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.textinput import TextInput
from kivy.uix.popup import Popup
from kivy.properties import BooleanProperty
from kivy.config import Config

normalfont = resource_filename(__name__,os.path.join('resources/fonts','DejaVuSans.ttf'))
boldfont = resource_filename(__name__,os.path.join('resources/fonts','DejaVuSans-Bold.ttf'))
fontawesomefont = resource_filename(__name__,os.path.join('resources/fonts','FontAwesome.ttf'))
buttonnormal = resource_filename(__name__,os.path.join('resources/images','button-normal.png'))
buttondown = resource_filename(__name__,os.path.join('resources/images','button-down.png'))
clearimage = resource_filename(__name__,os.path.join('resources/images','clear.png'))
backdrop = resource_filename(__name__,os.path.join('resources/images','backdrop.png'))
listbackdrop = resource_filename(__name__,os.path.join('resources/images','list-backdrop.png'))
trackslidercursor = resource_filename(__name__,os.path.join('resources/images','track-slider-cursor.png'))
ratingstars = [
    u"\uf006\uf006\uf006\uf006\uf006",
    u"\uf123\uf006\uf006\uf006\uf006",
    u"\uf005\uf006\uf006\uf006\uf006",
    u"\uf005\uf123\uf006\uf006\uf006",
    u"\uf005\uf005\uf006\uf006\uf006",
    u"\uf005\uf005\uf123\uf006\uf006",
    u"\uf005\uf005\uf005\uf006\uf006",
    u"\uf005\uf005\uf005\uf123\uf006",
    u"\uf005\uf005\uf005\uf005\uf006",
    u"\uf005\uf005\uf005\uf005\uf123",
    u"\uf005\uf005\uf005\uf005\uf005",
    u"\uf29c"
]
class ExtraSlider(Slider):
    """Class that implements some extra stuff on top of a standard slider."""

    def __init__(self,**kwargs):
        """Do normal init routine, but also register on_release event."""
        super(self.__class__,self).__init__(**kwargs)
        self.register_event_type('on_release')

    def on_release(self):
        """Override this with something you want this slider to do."""
        pass

    def on_touch_up(self, touch):
        """Check if slider is released, dispatch the on_release event if so."""
        released = super(self.__class__,self).on_touch_up(touch)
        if released:
            self.dispatch('on_release')
        return released

class OutlineLabel(Label):
    """A label that has an outline around it."""
    pass

class OutlineButton(Button,OutlineLabel):
    """A button with a label that has an outline around it."""
    pass

class ClearButton(Button,OutlineLabel):
    """A button that is clear instead of opaque."""
    pass

class OutlineTabbedPanelItem(TabbedPanelItem,OutlineLabel):
    """A label that has an outline around it."""
    pass

class InfoLargeLabel(OutlineLabel):
    """A label with large text."""
    pass

class ImageButton(ButtonBehavior, AsyncImage):
    """An image that you can press."""
    pass

class CoverButton(OutlineButton):
    img = ObjectProperty(None)
    layout = ObjectProperty(None)

    def __init__(self,**kwargs):
        super(self.__class__,self).__init__(**kwargs)
        self.background_normal = resource_filename(__name__,os.path.join('resources/images','clear.png'))
        self.background_down = resource_filename(__name__,os.path.join('resources/images','clear.png'))
        self.font_name = resource_filename(__name__,os.path.join('resources/fonts','DejaVuSans-Bold.ttf'))
        with self.canvas.before:
            Rectangle(texture=self.img.texture,pos=self.layout.pos,size=self.layout.size)

class ArtistRecycleBoxLayout(FocusBehavior,LayoutSelectionBehavior,RecycleBoxLayout):
    ''' Adds selection and focus behaviour to the view. '''

class ArtistRow(RecycleDataViewBehavior,BoxLayout):
    ''' Add selection support to the Label '''
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    def refresh_view_attrs(self, rv, index, data):
        ''' Catch and handle the view changes '''
        self.index = index
        return super(ArtistRow, self).refresh_view_attrs(
            rv, index, data)

    def on_touch_down(self, touch):
        ''' Add selection on touch down '''
        if super(ArtistRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        ''' Respond to the selection of items in the view. '''
        self.selected = is_selected
        if is_selected:
            App.get_running_app().root.selected_row=index

class UneditTextInput(TextInput):
    def insert_text(self, substring, from_undo=False):
        pass

class RatingPopup(Popup):
    rating_set = ObjectProperty(None)
    song = StringProperty(None)
    index = NumericProperty(None)

class StarBoxButton(BoxLayout,OutlineButton):
    rating = StringProperty(None)
    popup = ObjectProperty(None)

class CoverPopup(Popup):
    image_texture = ObjectProperty(None)



import os
import sys
from subprocess import call
from functools import partial

import kivy
kivy.require('1.10.0')
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.gridlayout import GridLayout
from kivy.uix.scrollview import ScrollView
from kivy.uix.popup import Popup
from kivy.uix.progressbar import ProgressBar
from kivy.uix.button import Button
from kivy.logger import Logger
from kivy.app import App
from kivy.factory import Factory
from kivy.lang.builder import Builder

from kmpc.widgets import OutlineTabbedPanelItem,OutlineLabel
from kmpc.sync import Sync,Subproc
from kmpc.kmpcapp import configdir

from twisted.internet.defer import Deferred,DeferredList

class GuiSync(Sync):

    def __init__(self,popup,config,runparts=[]):
        try:
            self.popup=popup
            self.pb={}
            App.get_running_app().root.do_idle_handler=False
            super(self.__class__,self).__init__(config,runparts)
        except Exception as e:
            Logger.error("__init__: "+format(e))

    def run_at_end(self,result):
        try:
            self.popup.dismiss()
            App.get_running_app().root.ids.system_tab.syncPopup.dismiss()
            App.get_running_app().root.do_idle_handler=True
        except Exception as e:
            Logger.error("run_at_end: "+format(e))

    def print_line(self,line):
        try:
            l=OutlineLabel(text=line.rstrip(),size_hint=(None,None),font_size='12sp',halign='left')
            l.bind(texture_size=l.setter('size'))
            self.popup.ids.layout.add_widget(l)
            self.popup.ids.sv.scroll_to(l)
            super(self.__class__,self).print_line(line)
        except Exception as e:
            Logger.error("print_line: "+format(e))

    def show_ratings_progress(self,done,total,sdir):
        try:
            self.pb[sdir].max=total
            self.pb[sdir].value=done
        except Exception as e:
            Logger.error("show_ratings_progress: "+format(e))

    def ratings_incoming(self,sdir):
        try:
            self.pb[sdir] = ProgressBar()
            l=OutlineLabel(text=sdir,size_hint=(None,None),font_size='12sp',halign='left')
            l.bind(texture_size=l.setter('size'))
            self.popup.ids.layout.add_widget(l)
            self.popup.ids.layout.add_widget(self.pb[sdir])
            self.popup.ids.sv.scroll_to(self.pb[sdir])
        except Exception as e:
            Logger.error("ratings_incoming: "+format(e))

class SystemTabbedPanelItem(OutlineTabbedPanelItem):
    """The System tab, for internal functions."""

    def __init__(self,**kwargs):
        super(self.__class__,self).__init__(**kwargs)
        self.syncPopup=Factory.SyncPopup()

    def sync_popup(self):
        self.syncPopup.open()

    def sync_fanart(self):
        stdoutPopup=Factory.StdoutPopup()
        stdoutPopup.ids.layout.bind(minimum_height=stdoutPopup.ids.layout.setter('height'))
        stdoutPopup.open()
        GuiSync(stdoutPopup,self.config,['fanart'])

    def sync_music(self):
        stdoutPopup=Factory.StdoutPopup()
        stdoutPopup.ids.layout.bind(minimum_height=stdoutPopup.ids.layout.setter('height'))
        stdoutPopup.open()
        GuiSync(stdoutPopup,self.config,['music'])

    def sync_export_ratings(self):
        stdoutPopup=Factory.StdoutPopup()
        stdoutPopup.ids.layout.bind(minimum_height=stdoutPopup.ids.layout.setter('height'))
        stdoutPopup.open()
        GuiSync(stdoutPopup,self.config,['exportratings'])

    def sync_import_ratings(self):
        stdoutPopup=Factory.StdoutPopup()
        stdoutPopup.ids.layout.bind(minimum_height=stdoutPopup.ids.layout.setter('height'))
        stdoutPopup.open()
        GuiSync(stdoutPopup,self.config,['importratings'])

    def sync_all(self):
        stdoutPopup=Factory.StdoutPopup()
        stdoutPopup.ids.layout.bind(minimum_height=stdoutPopup.ids.layout.setter('height'))
        stdoutPopup.open()
        if self.config.getboolean('system','exportfirst'):
            GuiSync(stdoutPopup,self.config,['music','fanart','exportratings','importratings'])
        else:
            GuiSync(stdoutPopup,self.config,['music','fanart','importratings','exportratings'])


    def update_print_line(self,popup,line):
        try:
            l=OutlineLabel(text=line.rstrip(),size_hint=(None,None),font_size='12sp',halign='left')
            l.bind(texture_size=l.setter('size'))
            popup.ids.layout.add_widget(l)
            popup.ids.sv.scroll_to(l)
            Logger.debug("Update: "+line.rstrip())
        except Exception as e:
            Logger.error("update_print_line: "+format(e))

    def update(self):
        """Runs the 'updatecommand' from the config file."""
        Logger.info('System: update')
        from twisted.internet import reactor
        updatePopup=Factory.StdoutPopup(title='Update kmpc')
        updatePopup.ids.layout.bind(minimum_height=updatePopup.ids.layout.setter('height'))
        updatePopup.open()
        d=Deferred()
        cb=[]
        cmdline=self.config.get('system','updatecommand').split(' ')
        pp=Subproc(partial(self.update_print_line,updatePopup))
        pp.deferred=Deferred()
        pp.deferred.addCallback(partial(self.closeit,updatePopup))
        reactor.spawnProcess(pp,cmdline[0],cmdline,{'PATH':os.environ['PATH']})

    def closeit(self,popup,r):
        popup.dismiss()

    def do_plugins(self):
        choosePluginPopup=Factory.ChoosePluginPopup()
        plugins=[]
        scripts=[]
        for dirpath, dirnames, filenames in os.walk(os.path.join(configdir,'plugins')):
            if 'plugin.kv' in filenames and 'plugin.py' in filenames:
                plugins.append(os.path.basename(dirpath))
            if 'plugin.sh' in filenames:
                scripts.append(os.path.basename(dirpath))
        for plugin in plugins:
            pbutton=Button(text=plugin)
            choosePluginPopup.ids.layout.add_widget(pbutton)
            pbutton.bind(on_press=choosePluginPopup.dismiss)
            pbutton.bind(on_press=partial(self.load_plugin,plugin))
        for script in scripts:
            pbutton=Button(text=script)
            choosePluginPopup.ids.layout.add_widget(pbutton)
            pbutton.bind(on_press=choosePluginPopup.dismiss)
            pbutton.bind(on_press=self.run_script)
            pbutton.script=os.path.join(configdir,'plugins',script,'plugin.sh')
        choosePluginPopup.open()

    def run_script(self,instance):
        Logger.debug("run_script: "+instance.script)
        call(instance.script,shell=True)

    def load_plugin(self,plugin,instance):
        Logger.debug("load_plugin: "+plugin)
        sys.path.append(os.path.join(configdir,'plugins',plugin))
        pluginmodule=__import__('plugin')
        reload(pluginmodule)
        Builder.load_file(os.path.join(configdir,'plugins',plugin,'plugin.kv'))
        pluginPopup=Factory.PluginPopup(title='Plugin: '+plugin)
        pluginContent=eval('Factory.'+plugin+'PluginContent()')
        pluginPopup.ids.plugincontent.add_widget(pluginContent)
        pluginPopup.ids.closebutton.bind(on_press=partial(self.unload_plugin,plugin,pluginmodule))
        pluginPopup.open()

    def unload_plugin(self,plugin,pluginmodule,instance):
        Logger.debug("unload_plugin: "+plugin)
        del pluginmodule
        Builder.unload_file(os.path.join(configdir,'plugins',plugin,'plugin.kv'))

    def do_reboot(self):
        """Method that reboots the host."""
        Logger.info('System: reboot')
        call(self.config.get('system','rebootcommand').split(' '))

    def do_poweroff(self):
        """Method that shuts down the host."""
        Logger.info('System: poweroff')
        call(self.config.get('system','poweroffcommand').split(' '))




import os
import io

import kivy
kivy.require('1.10.0')

from twisted.internet.defer import Deferred,DeferredList

from kivy.logger import Logger

class KmpcHelpers(object):

    def formatsong(self,rec):
        """Method used by library browser to properly format a song row."""
        song = ''
        # check if there is more than one disc and display if so
        dd=rec['disc'].split('/')
        if len(dd)>1:
            if int(dd[1]) > 1:
                song+='(Disc '+'%02d' % int(dd[0])+') '
        # sometimes track numbers are like '01/05' (one of five), so drop that second number
        tt=rec['track'].split('/')
        song+='%02d' % int(tt[0])+' '
        # if albumartist is different than track artist, display the track artist
        if rec['artist'] != rec['albumartist']:
            song+=rec['artist']+' - '
        # display the track title
        song+=rec['title']
        return song

    def getfontsize(self,str,scale=1):
        """Method that determines font size based on text length."""
        # helper array for scaling font sizes based on text length
        #sizearray = ['39sp','38sp','37sp','36sp','35sp','34sp','33sp','32sp','31sp','30sp','29sp','28sp','27sp','26sp','25sp']
        sizearray = ['39','38','37','36','35','34','33','32','31','30','29','28','27','26','25']
        lr = len(str)
        if lr < 33:
            rsize = '40'
        elif lr >= 55:
            rsize = '24'
        else:
            rsize = sizearray[int(round((lr-33)/21*14))]
        return format(int(round(int(rsize)/scale)))+'sp'

    def decodeFileName(self,name):
        """Method that tries to intelligently decode a filename to handle unicode weirdness."""
        if type(name) == str:
            try:
                name = name.decode('utf8')
            except:
                name = name.decode('windows-1252')
        return name

    def removeEmptyFolders(self,path,removeRoot=True):
        'Function to remove empty folders'
        if not os.path.isdir(path):
            return
        # remove empty subfolders
        files = os.listdir(path)
        if len(files):
            for f in files:
                fullpath = os.path.join(path, f)
                if os.path.isdir(fullpath):
                     self.removeEmptyFolders(fullpath)
        # if folder empty, delete it
        files = os.listdir(path)
        if len(files) == 0 and removeRoot:
            Logger.debug("removeEmptyDir:"+path)
            os.rmdir(path)




import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.logger import Logger
from kmpc.mpd import MPDProtocol
from twisted.internet import protocol

class MPDFactoryProtocol(MPDProtocol):
    """Factory for MPDProtocol."""
    def connectionMade(self):
        """Call the connectionMade override."""
        if callable(self.factory.connectionMade):
            self.factory.connectionMade(self)
    def connectionLost(self, reason):
        """Call the connectionLost override."""
        if callable(self.factory.connectionLost):
            self.factory.connectionLost(self, reason)

class MPDClientFactory(protocol.ClientFactory):
    """Factory for MPDClient."""
    protocol = MPDFactoryProtocol
    connectionMade = None
    connectionLost = None

    def __init__(self,idlehandler=None,**kwargs):
        self.idlehandler=idlehandler

    def buildProtocol(self, addr):
        """Hook up protocol and idle handler."""
        Logger.debug('MPDClientFactory: buildProtocol')
        protocol = self.protocol()
        protocol.factory = self
        if callable(self.idlehandler):
            protocol.idle_result = self.idlehandler
        return protocol

class MPDReconnectingClientFactory(protocol.ReconnectingClientFactory):
    """Factory for MPDClient."""
    protocol = MPDFactoryProtocol
    connectionMade = None
    connectionLost = None

    def __init__(self,idlehandler=None,**kwargs):
        self.idlehandler=idlehandler

    def buildProtocol(self, addr):
        """Hook up protocol and idle handler."""
        Logger.debug('MPDClientFactory: buildProtocol')
        protocol = self.protocol()
        protocol.factory = self
        if callable(self.idlehandler):
            protocol.idle_result = self.idlehandler
        return protocol

#    def clientConnectionFailed(self, connector, reason):
#        """mpd connection failed for some reason."""
#        Logger.error('Connection failed - goodbye!: '+format(reason))

#    def clientConnectionLost(self, connector, reason):
#        """mpd connection lost for some reason."""
#        Logger.error('Connection lost - goodbye!: '+format(reason))

# this class just returns a debug message for all calls to it to handle bad mpd connections
class Dummy(object):
    def __getattr__(self,attr):
        Logger.debug("MpdConnection: no connection when calling "+attr+" method")
        return self
    def __call__(self,*args):
        return self

class MpdConnection(object):

    def __init__(self,config,mpdhost,mpdport,idlehandler=None,initconnections=[],quiet=False,reconnect=True):
        self.config = config
        self.mpdhost = mpdhost
        self.mpdport = mpdport
        self.quiet = quiet
        Logger.debug("MpdConnection: connecting to "+mpdhost+":"+mpdport)
        # set up mpd connection
        self.initconnections=initconnections
        if reconnect: self.factory = MPDReconnectingClientFactory(idlehandler)
        else: self.factory = MPDClientFactory(idlehandler)
        self.factory.connectionMade = self.mpd_connectionMade
        self.factory.connectionLost = self.mpd_connectionLost
        from twisted.internet import reactor
        self.reactor=reactor.connectTCP(mpdhost, int(mpdport), self.factory)
        self.noprotocol=Dummy()

    # this part handles calls to protocol when it hasn't been set up yet or is incorrectly specified in config
    @property
    def protocol(self):
        try:
            if self.realprotocol:
                return self.realprotocol
        except AttributeError:
            Logger.debug("MpdConnection: no mpd connected")
            return self.noprotocol

    def mpd_connectionMade(self,protocol):
        """Callback when mpd is connected."""
        # copy the protocol to all the classes
        self.realprotocol = protocol
        if not self.quiet: Logger.info('MpdConnection: Connected to mpd server host='+self.mpdhost+' port='+self.mpdport)
        for ic in self.initconnections:
            if callable(ic):
                 ic(self)

    def mpd_connectionLost(self,protocol, reason):
        """Callback when mpd connection is lost."""
        if not self.quiet: Logger.warn('MpdConnection: Connection lost: %s' % reason)

    def handle_mpd_error(self,result):
        """Prints handled errors to the error log."""



import os
import pickle
from functools import partial
import musicbrainzngs
from PIL import Image as PImage
from ConfigParser import NoSectionError,NoOptionError

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

from kivy.app import App
from kivy.support import install_twisted_reactor
from kivy.logger import Logger
from kivy.network.urlrequest import UrlRequest
from kivy.uix.tabbedpanel import TabbedPanel
from kivy.clock import Clock

# import our local modules
from kmpc.version import VERSION, VERSION_STR
from kmpc.mpdfactory import MpdConnection
from kmpc.widgets import ArtistRecycleBoxLayout,ArtistRow,UneditTextInput

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

class ManagerInterface(TabbedPanel):

    def __init__(self,config,**kwargs):
        TabbedPanel.__init__(self,**kwargs)
        self.config=config
        self.artist_id_hash={}
        self.artist_name_hash={}
        self.media_hash={}
        self.wr_hash={}
        self.totaldone=0
        self.selected_row=None
        musicbrainzngs.set_useragent("kmpcmanager",VERSION_STR,'https://github.com/eratosthene/kmpc')
        self.fanarturl="http://webservice.fanart.tv/v3/music/"
        self.api_key="406b2a5af85c14b819c1c6332354b313"
        # install twisted reactor to interface with mpd
        install_twisted_reactor()
        # open mpd connection
        self.syncmpdconnection=MpdConnection(self.config,self.config.get('sync','synchost'),self.config.get('sync','syncmpdport'),None,[self.init_mpd])
        if not App.get_running_app().root:
            # root isn't defined yet, it must be us
            Logger.debug("ManagerInterface: running as full app")
        else:
            # root is something else
            Logger.debug("ManagerInterface: running as class")
            App.get_running_app().root.syncmpdconnection=self.syncmpdconnection

    def init_mpd(self,instance):
        self.refresh_artists_from_cache()

    def refresh_artists(self):
        self.syncmpdconnection.protocol.list('musicbrainz_artistid').addCallback(self.populate_artists).addErrback(self.syncmpdconnection.handle_mpd_error)

    def populate_artists(self,result):
        Logger.info("Manager: populate_artists")
        self.totaldone=0
        waittime=1
        self.wr_hash={}
        for row in result:
            aids = str(row)
            for aid in aids.split('/'):
                if aid not in self.artist_id_hash and aid not in self.wr_hash and len(aid)>0:
                    Logger.debug("MusicBrainz: scheduling query in "+str(waittime)+" seconds")
                    Clock.schedule_once(partial(self.query_mb,aid),waittime)
                    waittime=waittime+1
                    self.wr_hash[aid]=True

    def query_mb(self,aid,*largs):
        Logger.info("MusicBrainz: get_artist_by_id("+aid+")")
        try:
            mbres=musicbrainzngs.get_artist_by_id(aid)
        except WebServiceError as exc:
            Logger.error("MusicBrainz: web service error "+format(exc))
        else:
            aname=mbres['artist']['name']
            Logger.debug("query_mb: result from musicbrainz for aid "+aid+": "+aname)
            self.artist_id_hash[aid]=aname
            self.artist_name_hash[aname]=aid
            data = {'artist_id':aid,'artist_name':aname}
            self.ids.artist_tab.rv.data.append(data)
            self.ids.artist_tab.rv.refresh_from_data()
            self.totaldone=self.totaldone+1
            self.ids.status.text=aid+' ('+str(self.totaldone)+')'

    def write_artists_to_cache(self):
        cachefile=open(os.path.join(configdir,'artist_cache.pkl'),'w')
        pickle.dump((self.artist_id_hash,self.artist_name_hash,self.media_hash),cachefile,-1)
        cachefile.close()

    def refresh_artists_from_cache(self):
        try:
            cachefile=open(os.path.join(configdir,'artist_cache.pkl'),'r')
            (self.artist_id_hash,self.artist_name_hash,self.media_hash)=pickle.load(cachefile)
            cachefile.close()
        except IOError:
            pass
        self.ids.artist_tab.rv.data=[]
        newdata=[]
        for aid,aname in self.artist_id_hash.iteritems():
            try:
                has_artistbackground=self.media_hash[aid]['has_artistbackground']
            except:
                has_artistbackground=False
            try:
                has_logo=self.media_hash[aid]['has_logo']
            except:
                has_logo=False
            try:
                has_badge=self.media_hash[aid]['has_badge']
            except:
                has_badge=False
            datum = {'artist_id':aid,'artist_name':aname,'has_artistbackground':has_artistbackground,'has_logo':has_logo,'has_badge':has_badge}
            newdata.append(datum)
        self.ids.status.text='pulled '+str(len(self.artist_id_hash))+' lines from cache'
        self.ids.artist_tab.rv.data=sorted(newdata,key=lambda k: k['artist_name'])

    def scan_for_media(self,index):
        Logger.info('Manager: scanning '+self.ids.artist_tab.rv.data[index]['artist_id']+'for media')
        fa_path=self.config.get('paths','fanartpath')
        artistbackground_path=os.path.join(fa_path,self.ids.artist_tab.rv.data[index]['artist_id'],'artistbackground')
        logo_path=os.path.join(fa_path,self.ids.artist_tab.rv.data[index]['artist_id'],'logo')
        badge_path=os.path.join(fa_path,self.ids.artist_tab.rv.data[index]['artist_id'],'badge')
        self.ids.artist_tab.rv.data[index]['has_artistbackground']=os.path.isdir(artistbackground_path)
        self.ids.artist_tab.rv.data[index]['has_logo']=os.path.isdir(logo_path)
        self.ids.artist_tab.rv.data[index]['has_badge']=os.path.isdir(badge_path)
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]={}
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]['has_artistbackground']=os.path.isdir(artistbackground_path)
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]['has_logo']=os.path.isdir(logo_path)
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]['has_badge']=os.path.isdir(badge_path)
        self.ids.artist_tab.rv.refresh_from_data()

    def scan_row_for_media(self):
        if self.selected_row is not None:
            self.scan_for_media(self.selected_row)
        self.write_artists_to_cache()

    def scan_all_for_media(self,*args):
        for idx in range(0,len(self.ids.artist_tab.rv.data)):
            self.scan_for_media(idx)
        self.write_artists_to_cache()

    def trim_image(self,filename,request,result):
        Logger.debug("trim_image: fixing "+filename)
        image = PImage.open(filename)
        # convert to RGBa before getting bounding box to account for transparent pixels
        bbox = image.convert("RGBa").getbbox()
        # crop it and save
        image = image.crop(bbox)
        image.save(filename)

    def pull_art(self,index,*largs):
        Logger.info('Manager: pulling art for '+self.ids.artist_tab.rv.data[index]['artist_id'])
        aid=self.ids.artist_tab.rv.data[index]['artist_id']
        aname=self.ids.artist_tab.rv.data[index]['artist_name']
        fa_path=self.config.get('paths','fanartpath')
        fanart=self.fanarturl
        api_key=self.api_key
        furl=fanart+aid+"?api_key="+api_key
        client_key=self.config.get('fanart','client_key')
        if client_key:
            furl=furl+"&client_key="+client_key
        Logger.debug("pull_art: querying "+furl)
        request = UrlRequest(url=furl,on_success=partial(self.pull_art2,index))

    def pull_art2(self,index,request,result):
        aid=self.ids.artist_tab.rv.data[index]['artist_id']
        aname=self.ids.artist_tab.rv.data[index]['artist_name']
        fa_path=self.config.get('paths','fanartpath')
        d=result
        # see if there are blacklist entries for this artist
        bl=[]
        try:
            bl=self.config.get('artblacklist',aid).split(',')
        except NoSectionError:
            Logger.debug('pull_art2: no artblacklist section found')
        except NoOptionError:
            Logger.debug('pull_art2: no blacklist entries found for '+aid)
        except Exception as e:
            Logger.exception('pull_art2: '+format(e))
        else:
            Logger.debug('pull_art2: found blacklist entries for '+aid+': '+format(bl))
        if 'hdmusiclogo' in d or 'artistbackground' in d or 'musiclogo' in d:
            fapath=os.path.join(fa_path,aid)
            lpath=os.path.join(fapath,"logo")
            abpath=os.path.join(fapath,"artistbackground")
            bpath=os.path.join(fapath,"badge")
            try:
                Logger.debug("pull_art2: downloading to "+fapath)
                os.mkdir(fapath)
                with open(os.path.join(fapath,"__"+aname.replace(os.sep,'_')+"__"),'w'):
                    pass
            except OSError:
                pass
            if 'hdmusiclogo' in d:
                try:
                    os.mkdir(lpath)
                except OSError:
                    pass
                for idx,img in enumerate(d['hdmusiclogo']):
                    if not os.path.isfile(os.path.join(lpath,img['id']+'.png')) and not os.path.isfile(os.path.join(bpath,img['id']+'.png')) and not img['id'] in bl:
                        Logger.debug("pull_art2: downloading hdmusiclogo "+img['id'])
                        fp=os.path.join(lpath,img['id']+'.png')
                        req = UrlRequest(img['url'],on_success=partial(self.trim_image,fp),file_path=fp)
                        if self.config.getboolean('logs','artlog'):
                            adfile=open(os.path.join(configdir,'artlog.txt'),'a')
                            adfile.write(os.path.join(lpath,img['id']+'.png')+"\n")
                            adfile.close()
            if 'musiclogo' in d:
                try:
                    os.mkdir(lpath)
                except OSError:
                    pass
                for idx,img in enumerate(d['musiclogo']):
                    if not os.path.isfile(os.path.join(lpath,img['id']+'.png')) and not os.path.isfile(os.path.join(bpath,img['id']+'.png')) and not img['id'] in bl:
                        Logger.debug("pull_art2: downloading musiclogo "+img['id'])
                        fp=os.path.join(lpath,img['id']+'.png')
                        req = UrlRequest(img['url'],on_success=partial(self.trim_image,fp),file_path=fp)
                        if self.config.getboolean('logs','artlog'):
                            adfile=open(os.path.join(configdir,'artlog.txt'),'a')
                            adfile.write(os.path.join(lpath,img['id']+'.png')+"\n")
                            adfile.close()
            if 'artistbackground' in d:
                try:
                    os.mkdir(abpath)
                except OSError:
                    pass
                for idx,img in enumerate(d['artistbackground']):
                    if not os.path.isfile(os.path.join(abpath,img['id']+'.png')) and not img['id'] in bl:
                        Logger.debug("pull_art2: downloading artistbackground "+img['id'])
                        fp=os.path.join(abpath,img['id']+'.png')
                        req = UrlRequest(img['url'],file_path=fp)
                        if self.config.getboolean('logs','artlog'):
                            adfile=open(os.path.join(configdir,'artlog.txt'),'a')
                            adfile.write(os.path.join(abpath,img['id']+'.png')+"\n")
                            adfile.close()

    def pull_art_for_row(self):
        if self.selected_row is not None:
            self.pull_art(self.selected_row)
            self.scan_for_media(self.selected_row)
        self.write_artists_to_cache()

    def pull_art_for_all(self):
        waittime=1
        for idx in range(0,len(self.ids.artist_tab.rv.data)):
            Logger.debug("FanArt.tv: scheduling query in "+str(waittime)+" seconds")
            Clock.schedule_once(partial(self.pull_art,idx),waittime)
            waittime=waittime+1
        Clock.schedule_once(self.scan_all_for_media,waittime+5)
        self.write_artists_to_cache()




import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.logger import Logger
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.properties import BooleanProperty,StringProperty
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.popup import Popup
from kivy.uix.textinput import TextInput
from kivy.uix.button import Button
from kivy.uix.vkeyboard import VKeyboard
from kivy.factory import Factory
from kivy.clock import Clock
from functools import partial

from kmpc.widgets import OutlineTabbedPanelItem

class PlaylistTabbedPanelItem(OutlineTabbedPanelItem):
    """The Playlist tab, shows the current playlist and allows interacting with it."""
    playlist_selection={}

    def playlist_clear_pressed(self):
        """Callback for playlist clear button."""
        Logger.info("Playlist: clear")
        App.get_running_app().root.mpdconnection.protocol.clear()

    def playlist_delete_pressed(self):
        """Callback for playlist delete button."""
        Logger.info("Playlist: delete")
        # loop through all selected tracks and remove each one from the playlist
        for pos in self.playlist_selection:
            songid=str(self.rv.data[pos]['songid'])
            Logger.debug("Playlist: deleting songid "+songid)
            App.get_running_app().root.mpdconnection.protocol.deleteid(songid)
        self.rbl.clear_selection()

    def playlist_move_pressed(self):
        """Callback for playlist move button."""
        # TODO: implement this
        Logger.info("Playlist: move")
        Logger.warn("Playlist: move not implemented")
        self.rbl.clear_selection()

    def playlist_shuffle_pressed(self):
        """Callback for playlist shuffle button."""
        Logger.info("Playlist: shuffle")
        # shuffle the playlist. note that this is different from toggling random playback, as it
        # actually reorders the playlist randomly rather than just playing in random order
        App.get_running_app().root.mpdconnection.protocol.shuffle()
        self.rbl.clear_selection()

    def playlist_swap_pressed(self):
        """Callback for playlist swap button."""
        Logger.info("Playlist: swap")
        # if exactly 2 tracks are selected, swap them
        if len(self.playlist_selection) != 2:
            Logger.warn("Playlist: swap only works with two rows selected")
        else:
            s1 = self.playlist_selection.keys()[0]
            s2 = self.playlist_selection.keys()[1]
            App.get_running_app().root.mpdconnection.protocol.swap(str(s1),str(s2)).addErrback(self.handle_mpd_error)
        self.rbl.clear_selection()

    def playlist_save_pressed(self):
        """Callback for playlist save button."""
        Logger.info("Playlist: save")
        self.rbl.clear_selection()
        popup=Factory.PlaylistSavePopup()
        popup.open()

    def save_playlist(self,t,popup):
        """Tell mpd to save the current playlist with the name that was input."""
        Logger.info("Playlist: save_playlist("+t+")")
        App.get_running_app().root.mpdconnection.protocol.save(t).addErrback(self.handle_mpd_error)
        popup.dismiss()

    def populate_playlist(self,result):
        """Callback for mpd playlist info."""
        Logger.info("Playlist: populate_playlist()")
        self.rv.data = []
        # loop through mpd playlist info and add to the recycleview
        for row in result:
            Logger.debug("Playlist: row "+row['pos']+" found = "+row['title'])
            r = {'plpos':row['pos'],'rownum':str(int(row['pos'])+1),'artist':format(row['artist']),'title':format(row['title']),'songid':format(row['id'])}
            self.rv.data.append(r)
        # when playlist is populated, also ask mpd for current status to highlight current track
        App.get_running_app().root.mpdconnection.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)

    def update_mpd_status(self,result):
        """Callback for mpd status about current track."""
        Logger.debug('Playlist: update_mpd_status()')
        # TODO: pretty sure this is what crashes the player on super-long playlists
        if result['state'] == 'stop':
            # if stopped, there's no current track
            self.currsong=None
            # loop through recycleview, unhighlight all tracks
            for r in self.rv.data:
                r['iscurrent'] = False
        else:
            self.currsong=result['song']
            # loop through recycleview, highlight current track and unhighlight all others
            for r in self.rv.data:
                if r['plpos'] == result['song']:
                    r['iscurrent'] = True
                else:
                    r['iscurrent'] = False
        # tell the recycleview to refresh itself from the changed data
        self.rv.refresh_from_layout()

    def handle_mpd_error(self,result):
        """Callback for handling mpd exceptions."""
        Logger.error('Playlist: MPDIdleHandler Callback error: {}'.format(result))

class PlaylistRecycleBoxLayout(LayoutSelectionBehavior,RecycleBoxLayout):
    """Adds selection and focus behaviour to the recyclebox."""

class PlaylistRow(RecycleDataViewBehavior,BoxLayout):
    """Adds selection and long-press support to the recycleview."""
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    def playfrom(self, touch, index, *args):
        """Handle long-press on a playlist row."""
        Logger.debug("Playlist: long-touch playfrom "+str(index))
        App.get_running_app().root.mpdconnection.protocol.play(str(index))
        App.get_running_app().root.ids.playlist_tab.rbl.clear_selection()

    def refresh_view_attrs(self, rv, index, data):
        """Catch and handle the view changes."""
        self.index = index
        return super(self.__class__, self).refresh_view_attrs(
            rv, index, data)

    def on_touch_down(self, touch):
        """Adds selection and long-press on touch down."""
        if super(self.__class__, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            # these lines start a 1 second clock to detect long-presses
            callback = partial(self.playfrom, touch, self.index)
            Clock.schedule_once(callback, 1)
            touch.ud['event'] = callback
            return self.parent.select_with_touch(self.index, touch)

    def on_touch_up(self, touch):
        """Clean up long-press on touch up."""
        if super(PlaylistRow, self).on_touch_up(touch):
            return True
        # if i don't check for this, the app crashes when things scroll off screen
        if 'event' in touch.ud:
            Clock.unschedule(touch.ud['event'])

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        self.selected = is_selected
        pt=App.get_running_app().root.ids.playlist_tab
        if is_selected:
            pt.playlist_selection[index] = True
        else:
            if index in pt.playlist_selection:
                del pt.playlist_selection[index]




# import dependencies
import os
import sys
from pkg_resources import resource_filename

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

# import all the other kivy stuff
from kivy.config import Config,ConfigParser
from kivy.app import App
from kivy.logger import Logger
from kivy.lang import Builder
from kivy.utils import get_color_from_hex

# import our local modules
from kmpc.version import VERSION_STR
from kmpc.sync import Sync
from kmpc.kmpcinterface import KmpcInterface

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

# load the interface kv files
Builder.load_file(resource_filename(__name__,os.path.join('resources/kv','widgets.kv')))
Builder.load_file(resource_filename(__name__,os.path.join('resources/kv','library.kv')))
Builder.load_file(resource_filename(__name__,os.path.join('resources/kv','playlist.kv')))
Builder.load_file(resource_filename(__name__,os.path.join('resources/kv','system.kv')))
Builder.load_file(resource_filename(__name__,os.path.join('resources/kv','interface.kv')))

class KmpcApp(App):
    """The overall app class, builds the main interface widget."""

    def __init__(self,args):
        """Override kivy config values with necessary ones"""
        self.args=args
        super(self.__class__,self).__init__()

    def build_config(self,config):
        config.setdefaults('mpd', {
            'mpdhost': '127.0.0.1',
            'mpdport': '6600'
        })
        config.setdefaults('paths', {
            'musicpath': '/mnt/music',
            'fanartpath': '/mnt/fanart',
            'tmppath': '/tmp'
        })
        config.setdefaults('sync', {
            'synchost': '127.0.0.1',
            'syncmpdport': '6600',
            'syncmusicpath': '/mnt/music',
            'syncfanartpath': '/mnt/fanart',
            'synctmppath': '/tmp',
            'syncplaylist': 'synclist'
        })
        config.setdefaults('system', {
            'rpienable': '0',
            'originalyear': '1',
            'advancedtitles': '0',
            'updatecommand': 'sudo pip install -U kmpc --no-deps',
            'rebootcommand': 'sudo reboot',
            'poweroffcommand': 'sudo poweroff',
            'exportfirst': '1',
        })
        config.setdefaults('songratings', {
            'star0': 'Silence',
            'star1': 'Songs that should never be heard',
            'star2': 'Songs no one likes',
            'star3': 'Songs for certain occasions',
            'star4': 'Songs someone else likes',
            'star5': 'Filler tracks with no music',
            'star6': 'Meh track or short musical filler',
            'star7': 'Occasional listening songs',
            'star8': 'Great songs for all occasions',
            'star9': 'Best songs by an artist',
            'star10': 'Favorite songs of all time'
        })
        config.setdefaults('colors', {
            'button': '#00B361',
            'backdrop': '#4096FF',
            'listitem': '#4080FF',
            'listitemselected': '#FFFF00',
            'listitemcurrent': '#521C4F'
        })

    def get_color(self,c):
        cc=self.config.get('colors',c)
        try:
            # if backdrop or button, alpha channel=1, else 0.5
            if c in ['backdrop','button']: cc=cc+'FF'
            else: cc=cc+'80'
            t=get_color_from_hex(cc)
        except:
            Logger.error("Application: color "+cc+" for "+c+" is invalid")
            t=(1,1,1,1)
        return t

    def get_application_config(self):
        return super(self.__class__,self).get_application_config(configdir+'/config.ini')

    def build_settings(self,settings):
        settings.add_json_panel('mpd settings',self.config,resource_filename(__name__,os.path.join('resources/json','config_mpd.json')))
        settings.add_json_panel('path settings',self.config,resource_filename(__name__,os.path.join('resources/json','config_paths.json')))
        settings.add_json_panel('sync settings',self.config,resource_filename(__name__,os.path.join('resources/json','config_sync.json')))
        settings.add_json_panel('system settings',self.config,resource_filename(__name__,os.path.join('resources/json','config_system.json')))
        settings.add_json_panel('song ratings',self.config,resource_filename(__name__,os.path.join('resources/json','config_star.json')))
        settings.add_json_panel('colors',self.config,resource_filename(__name__,os.path.join('resources/json','config_colors.json')))

    def on_config_change(self, config, section, key, value):
        if config is self.config:
            Logger.info("Application: config entry has changed: ["+section+"] "+key+"="+value)
            if callable (self.root.settings_update): self.root.settings_update()

    def build(self,*args):
        """Instantiates KmpcInterface."""
        self.version_str=VERSION_STR
        if not os.path.isdir(configdir):
            os.mkdir(configdir)
        # try to read existing config file
        self.config=self.load_config()
        # write out config file in case it doesn't exist yet
        self.config.write()
        if self.args.newconfig:
            sys.exit(0)
        elif self.args.sync:
            if self.args.sync=='all':
                if self.config.getboolean('system','exportfirst'):
                    s=Sync(self.config,['music','fanart','exportratings','importratings'])
                else:
                    s=Sync(self.config,['music','fanart','importratings','exportratings'])
            else:
                s=Sync(self.config,[self.args.sync])
            sys.exit(0)
        else:
            return KmpcInterface(self.config)

if __name__ == '__main__':
    # run the app!
    KmpcApp().run()




from copy import deepcopy
from functools import partial
import os

import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.logger import Logger
from twisted.internet.defer import inlineCallbacks
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.properties import BooleanProperty,ObjectProperty
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.popup import Popup
from kivy.factory import Factory
from kivy.clock import Clock

from kmpc.extra import KmpcHelpers
from kmpc.widgets import fontawesomefont
from kmpc.librarypanel import LibraryTabbedPanelItem,LibraryRow

Helpers=KmpcHelpers()

class ManagerLibraryTabbedPanelItem(LibraryTabbedPanelItem):

    def render_row(self,r,has_sticker,result):
        rr=deepcopy(r)
        if has_sticker:
            rr['copy_flag']=str(result)
        else:
            rr['copy_flag']=''
        App.get_running_app().root.syncmpdconnection.protocol.sticker_get('song',rr['base'],'rating').addCallback(partial(self.render_row2,rr,True)).addErrback(partial(self.render_row2,rr,False))

    def render_row2(self,r,has_sticker,result):
        rr=deepcopy(r)
        if has_sticker:
            rr['rating']=str(result)
        else:
            rr['rating']=''
        self.rv.data.append(rr)

    def reload_view(self,result):
        Logger.info("Library: reload_view() current type: "+self.current_view['info']['type'])
        self.rv.data=[]
        if self.current_view['info']['type'] == 'uri':
            if self.current_view['base'] != '/':
                (hbase,tbase)=os.path.split(self.current_view['base'])
                (b1,b2)=os.path.split(hbase)
                if b2 == '':
                    b2 = 'root'
                upbase=os.path.normpath(self.current_view['base']+os.sep+'..')
                if upbase == '.':
                    upbase = '/'
                r = {'value':"up to "+b2,'base':upbase,'info':{'type':'uri'},'copy_flag':'','rating':''}
                self.rv.data.append(r)
                self.current_header.text = tbase
            else:
                self.current_header.text = 'All Files'
        elif self.current_view['info']['type'] == 'albumartistsort':
            r={'value': 'up to All Album Artists','base':'All Album Artists','info':{'type':'rootalbums'},'copy_flag':'','rating':''}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        elif self.current_view['info']['type'] == 'album':
            r={'value':'up to '+self.current_view['info']['albumartistsort'],'base':self.current_view['info']['albumartistsort'],'info':{'type':'albumartistsort'},'copy_flag':'','rating':''}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        elif self.current_view['info']['type'] == 'artistsort':
            r={'value': 'up to All Track Artists','base':'All Track Artists','info':{'type':'roottracks'},'copy_flag':'','rating':''}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        else:
            self.current_header.text = self.current_view['base']
        for row in result:
            if 'playlist' in row:
                if self.current_view['info']['type'] != 'uri':
                    Logger.debug("Library: playlist found = "+row['playlist'])
                    r = {'value':row['playlist'],'base':row['playlist'],'info':{'type':'playlist'},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
            elif 'directory' in row:
                Logger.debug("Library: directory found: ["+row['directory']+"]")
                (b1,b2)=os.path.split(row['directory'])
                r={'value':b2,'base':row['directory'],'info':{'type':'uri'},'copy_flag':'','rating':''}
                self.rv.data.append(r)
            elif 'file' in row:
                Logger.debug("FileBrowser: file found: ["+row['file']+"]")
                r={'value':Helpers.formatsong(row),'base':row['file'],'info':{'type':'file'}}
                App.get_running_app().root.syncmpdconnection.protocol.sticker_get('song',row['file'],'copy_flag').addCallback(partial(self.render_row,r,True)).addErrback(partial(self.render_row,r,False))
            else:
                if self.current_view['info']['type'] == 'rootalbums':
                    Logger.debug("Library: album artist found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'albumartistsort'},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'albumartistsort':
                    Logger.debug("Library: album found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'album','albumartistsort':self.current_view['base']},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'roottracks':
                    Logger.debug("Library: track artist found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'artistsort'},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'artistsort':
                    Logger.debug("Library: track found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'track','artistsort':self.current_view['base']},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                else:
                    Logger.warn("Library: not sure what to do with ["+format(row)+"]")

    def handle_double_click(self,row,index):
        Logger.debug("Library: handle_double_click("+format(row)+")")
        self.current_view = deepcopy(row)
        if row['info']['type'] == 'uri':
            App.get_running_app().root.syncmpdconnection.protocol.lsinfo(row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'rootalbums':
            App.get_running_app().root.syncmpdconnection.protocol.list('albumartistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'albumartistsort':
            App.get_running_app().root.syncmpdconnection.protocol.list('album','albumartistsort',row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'album':
            App.get_running_app().root.syncmpdconnection.protocol.find('album',row['base'],'albumartistsort',row['info']['albumartistsort']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'roottracks':
            App.get_running_app().root.syncmpdconnection.protocol.list('artistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'artistsort':
            App.get_running_app().root.syncmpdconnection.protocol.list('title','artistsort',row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'playlist':
            pass
        elif row['info']['type'] == 'file':
            pass
        else:
            Logger.warn("Library: double click for ["+format(row)+"] not implemented")

    def reload_row_after_sticker(self,copy_flag,index,result):
        self.rv.data[index]['copy_flag']=copy_flag
        self.rv.refresh_from_data()

    def set_copy_flag_find(self,copy_flag,index,result):
        for rrow in result:
            if copy_flag:
                Logger.debug("set_copy_flag_find: setting copy_flag to "+copy_flag+" for file "+rrow['file'])
                App.get_running_app().root.syncmpdconnection.protocol.sticker_set('song',rrow['file'],'copy_flag',copy_flag).addCallback(partial(self.reload_row_after_sticker,copy_flag,index)).addErrback(self.handle_mpd_error)
            else:
                Logger.debug("set_copy_flag_find: clearing copy_flag for file "+rrow['file'])
                App.get_running_app().root.syncmpdconnection.protocol.sticker_delete('song',rrow['file'],'copy_flag').addCallback(partial(self.reload_row_after_sticker,'',index)).addErrback(self.handle_mpd_error)

    def set_copy_flag_find_one(self,copy_flag,index,result):
        for rrow in result:
            if copy_flag:
                Logger.debug("set_copy_flag_find_one: setting copy_flag to "+copy_flag+" for file "+rrow['file'])
                App.get_running_app().root.syncmpdconnection.protocol.sticker_set('song',rrow['file'],'copy_flag',copy_flag).addCallback(partial(self.reload_row_after_sticker,copy_flag,index)).addErrback(self.handle_mpd_error)
            else:
                Logger.debug("set_copy_flag_find_one: clearing copy_flag for file "+rrow['file'])
                App.get_running_app().root.syncmpdconnection.protocol.sticker_delete('song',rrow['file'],'copy_flag').addCallback(partial(self.reload_row_after_sticker,'',index)).addErrback(self.handle_mpd_error)
            break

    def set_copy_flag(self,copy_flag):
        Logger.info('Library: set_copy_flag('+str(copy_flag)+')')
        for index in self.rbl.selected_nodes:
            row = self.rv.data[index]
            mtype=row['info']['type']
            Logger.info("Library: Setting copy_flag for "+mtype+" '"+row['base']+"' to "+copy_flag)
            if mtype == 'file':
                Logger.debug("set_copy_flag: adding uri or file")
                if copy_flag:
                    Logger.debug("set_copy_flag: setting copy_flag to "+copy_flag+" for file "+row['base'])
                    App.get_running_app().root.syncmpdconnection.protocol.sticker_set('song',row['base'],'copy_flag',copy_flag).addCallback(partial(self.reload_row_after_sticker,copy_flag,index)).addErrback(self.handle_mpd_error)
                else:
                    Logger.debug("set_copy_flag: clearing copy_flag for file "+row['base'])
                    App.get_running_app().root.syncmpdconnection.protocol.sticker_delete('song',row['base'],'copy_flag').addCallback(partial(self.reload_row_after_sticker,'',index)).addErrback(self.handle_mpd_error)
            elif mtype == 'albumartistsort':
                App.get_running_app().root.syncmpdconnection.protocol.find(mtype,row['base']).addCallback(partial(self.set_copy_flag_find,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'album':
                App.get_running_app().root.syncmpdconnection.protocol.find(mtype,row['base'],'albumartistsort',row['info']['albumartistsort']).addCallback(partial(self.set_copy_flag_find,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'artistsort':
                App.get_running_app().root.syncmpdconnection.protocol.find(mtype,row['base']).addCallback(partial(self.set_copy_flag_find,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'track':
                App.get_running_app().root.syncmpdconnection.protocol.find('artistsort',row['info']['artistsort'],'title',row['base']).addCallback(partial(self.set_copy_flag_find_one,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'playlist' or mtype == 'uri':
                Logger.info("Library: "+mtype+" copy_flag not implemented")
                #App.get_running_app().root.syncmpdconnection.protocol.load(row['base'])
            else:
                Logger.warning("Library: "+mtype+' copy_flag not implemented')
        self.rbl.clear_selection()

    def rating_set(self,index,rating,popup):
        Logger.debug('Application: rating_set('+rating+')')
        popup.dismiss()
        if rating:
            App.get_running_app().root.syncmpdconnection.protocol.sticker_set('song',self.rv.data[index]['base'],'rating',rating).addCallback(partial(self.handle_rating_set,index,rating,True)).addErrback(partial(self.handle_rating_set,index,rating,False))
        else:
            App.get_running_app().root.syncmpdconnection.protocol.sticker_delete('song',self.rv.data[index]['base'],'rating').addCallback(partial(self.handle_rating_set,index,rating,True)).addErrback(partial(self.handle_rating_set,index,rating,False))

    def handle_rating_set(self,index,rating,succ,result):
        if succ:
            Logger.info("Library: successfully set song rating for "+self.rv.data[index]['base'])
            self.rv.data[index]['rating']=rating
            self.rv.refresh_from_data()
        else:
            Logger.info("Library: could not set song rating for "+self.rv.data[index]['base'])

    def generate_list(self,ltype,minstars,op='>=',pname='playlist'):
        Logger.info('generate_list: generating with minimum stars '+str(minstars))
        self.tlist={}
        App.get_running_app().root.syncmpdconnection.protocol.sticker_find('song','','rating').addCallback(partial(self.generate_play_list,ltype,minstars,op,pname)).addErrback(self.handle_mpd_error)

    def generate_play_list(self,ltype,minstars,op,pname,result):
        Logger.debug("generate_play_list: "+ltype)
        for row in result:
            rating=row['sticker'].split('=')[1]
            uri=row['file']
            if (op=='<' and int(rating)<int(minstars)) or (op=='<=' and int(rating)<=int(minstars)) or (op=='=' and int(rating)==int(minstars)) or (op=='>=' and int(rating)>=int(minstars)) or (op=='>' and int(rating)>int(minstars)):
                Logger.debug("generate_play_list: rating ["+rating+"] file ["+uri+"]")
                self.tlist[uri]=1
        if ltype=='playlist':
            Logger.info("generate_play_list: writing to playlist ["+pname+"]")
            App.get_running_app().root.syncmpdconnection.protocol.playlistclear(pname).addErrback(self.handle_mpd_error)
            for k in sorted(self.tlist.keys()):
                App.get_running_app().root.syncmpdconnection.protocol.playlistadd(pname,k).addErrback(self.handle_mpd_error)
        elif ltype=='synclist':
            App.get_running_app().root.syncmpdconnection.protocol.sticker_find('song','','copy_flag').addCallback(self.generate_sync_list).addErrback(self.handle_mpd_error)

    def generate_sync_list(self,result):
        for row in result:
            uri=row['file']
            if row['sticker'] == 'copy_flag=Y':
                Logger.debug("generate_sync_list: copy flag adding "+uri)
                self.tlist[uri]=1
            else:
                Logger.debug("generate_sync_list: copy flag removing "+uri)
                try:
                    del self.tlist[uri]
                except KeyError:
                    pass
        Logger.info("generate_sync_list: writing to playlist ["+App.get_running_app().root.config.get('sync','syncplaylist')+"]")
        App.get_running_app().root.syncmpdconnection.protocol.playlistclear(App.get_running_app().root.config.get('sync','syncplaylist')).addErrback(self.handle_mpd_error)
        for k in sorted(self.tlist.keys()):
            App.get_running_app().root.syncmpdconnection.protocol.playlistadd(App.get_running_app().root.config.get('sync','syncplaylist'),k).addErrback(self.handle_mpd_error)

    def change_view_type(self,value):
        Logger.info("Library: View changed to "+value)
        self.rbl.clear_selection()
        if value == 'Files':
            self.current_view = {'value': 'root','base':'/','info':{'type':'uri'}}
            App.get_running_app().root.syncmpdconnection.protocol.lsinfo(self.current_view['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='down'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='normal'
        elif value == 'Albums':
            self.current_view = {'value': 'All Album Artists','base':'All Album Artists','info':{'type':'rootalbums'}}
            App.get_running_app().root.syncmpdconnection.protocol.list('albumartistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='down'
            self.ids.tracks_button.state='normal'
        elif value == 'Tracks':
            self.current_view = {'value': 'All Track Artists','base':'All Track Artists','info':{'type':'roottracks'}}
            App.get_running_app().root.syncmpdconnection.protocol.list('artistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='down'

    def popup_generate(self):
        """Callback when user presses the Generate button."""
        generatePopup=Factory.ManagerGeneratePopup(library_tab=self)
        generatePopup.open()

class ManagerGeneratePopup(Popup):
    library_tab=ObjectProperty(None)

class ManagerLibraryRow(LibraryRow):
    ''' Add selection support to the Label '''

    def rating_popup(self,instance):
        Logger.debug('Library: rating_popup()')
        popup=Factory.RatingPopup(index=self.index,rating_set=App.get_running_app().root.ids.library_tab.rating_set)
        popup.open()

    def long_touch(self, touch, index, *args):
        """Callback when user long-presses on a row."""
        Logger.debug("Library: long-touch on "+str(index))
        App.get_running_app().root.ids.library_tab.rbl.clear_selection()
        App.get_running_app().root.ids.library_tab.handle_double_click(App.get_running_app().root.ids.library_tab.rv.data[index],index)

    def on_touch_down(self, touch):
        """Adds selection, long-press handling on touch down."""
        if BoxLayout.on_touch_down(self,touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            # these lines start a 1 second clock to detect long-presses
            callback = partial(self.long_touch, touch, self.index)
            Clock.schedule_once(callback, 1)
            touch.ud['event'] = callback
            if touch.is_double_tap:
                Logger.debug("Library: double-click on "+str(self.index))
                App.get_running_app().root.ids.library_tab.rbl.clear_selection()
                App.get_running_app().root.ids.library_tab.handle_double_click(App.get_running_app().root.ids.library_tab.rv.data[self.index],self.index)
            else:
                return self.parent.select_with_touch(self.index, touch)
